#!/usr/bin/env python

# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import json
import re
import sys
import time
import webbrowser
from contextlib import nullcontext
from enum import Enum
from logging import getLogger
from pathlib import Path
from types import SimpleNamespace
from typing import Annotated, Any, NoReturn

import click
import typer
import yaml
from jinja2 import Environment, PackageLoader, StrictUndefined
from pydantic import ValidationError as PydanticValidationError
from rich.console import Console as RichConsole
from rich.table import Table as RichTable

from . import engine
from .api_parse import (
    EXPECTED_OBJECTS,
    TesseractBuildConfig,
    TesseractConfig,
    ValidationError,
    get_non_base_fields_in_tesseract_config,
)
from .config import get_config
from .docker_client import (
    APIError,
    BuildError,
    CLIDockerClient,
    Container,
    ContainerError,
    Image,
    ImageNotFound,
    NotFound,
)
from .exceptions import UserError
from .logs import DEFAULT_CONSOLE, set_logger

logger = getLogger("tesseract")

# Jinja2 Template Environment
ENV = Environment(
    loader=PackageLoader("tesseract_core.sdk", "templates"),
    undefined=StrictUndefined,
)

docker_client = CLIDockerClient()


class SpellcheckedTyperGroup(typer.core.TyperGroup):
    """A Typer group that suggests similar commands if a command is not found."""

    def get_command(self, ctx: click.Context, invoked_command: str) -> Any:
        """Get a command from the Typer group, suggesting similar commands if the command is not found."""
        import difflib

        possible_commands = self.list_commands(ctx)
        if invoked_command not in possible_commands:
            close_match = difflib.get_close_matches(
                invoked_command, possible_commands, n=1, cutoff=0.6
            )
            if close_match:
                raise click.UsageError(
                    f"No such command '{invoked_command}'. Did you mean '{close_match[0]}'?",
                    ctx,
                )
        return super().get_command(ctx, invoked_command)


app = typer.Typer(
    # Make -h an alias for --help
    context_settings={"help_option_names": ["-h", "--help"]},
    name="Tesseract",
    pretty_exceptions_show_locals=False,
    cls=SpellcheckedTyperGroup,
)

# Module-wide config
state = SimpleNamespace()
state.print_user_error_tracebacks = False

# Create a list of possible commands based on the ones in api_parse (kebab-cased)
POSSIBLE_CMDS = set(
    re.sub(r"([a-z])([A-Z])", r"\1-\2", object.name).replace("_", "-").lower()
    for object in EXPECTED_OBJECTS
)
POSSIBLE_CMDS.update({"health", "openapi-schema", "check", "check-gradients"})

# All fields in TesseractConfig and TesseractBuildConfig for config override
POSSIBLE_KEYPATHS = TesseractConfig.model_fields.keys()
# Check that the only field that has nested fields is build_config
assert len(get_non_base_fields_in_tesseract_config()) == 1
POSSIBLE_BUILD_CONFIGS = TesseractBuildConfig.model_fields.keys()

# Traverse templates folder to seach for recipes
AVAILABLE_RECIPES = set()
for temp_with_path in ENV.list_templates(extensions=["py"]):
    temp_with_path = Path(temp_with_path)
    if temp_with_path.name == "tesseract_api.py" and temp_with_path.parent:
        AVAILABLE_RECIPES.add(str(temp_with_path.parent))
AVAILABLE_RECIPES = sorted(AVAILABLE_RECIPES)


class LogLevel(str, Enum):
    """Available log levels for Tesseract CLI."""

    # Must be an enum to represent a choice in Typer
    debug = "debug"
    info = "info"
    warning = "warning"
    error = "error"
    critical = "critical"


def _validate_tesseract_name(name: str | None) -> str:
    if name is None:
        if sys.stdout.isatty() and sys.stdin.isatty():
            name = typer.prompt("Enter a name for the Tesseract")
        else:
            raise typer.BadParameter(
                "Name must be provided as an argument or interactively."
            )

    forbidden_characters: str = ":;,.@#$%^&*()[]{}<>?|\\`~"
    if any(char in forbidden_characters for char in name) or any(
        char.isspace() for char in name
    ):
        raise typer.BadParameter(
            f"Name cannot contain whitespace or any of the following characters: {forbidden_characters}"
        )
    return name


def version_callback(value: bool | None) -> None:
    """Typer callback for version option."""
    if value:
        from tesseract_core import __version__

        typer.echo(__version__)
        raise typer.Exit()


@app.callback()
def main_callback(
    loglevel: Annotated[
        LogLevel,
        typer.Option(
            help="Set the logging level. At debug level, also print tracebacks for user errors.",
            case_sensitive=False,
            show_default=True,
            metavar="LEVEL",
        ),
    ] = LogLevel.info,
    version: Annotated[
        bool | None,
        typer.Option(
            "--version",
            callback=version_callback,
            is_eager=True,
            help="Print Tesseract CLI version and exit.",
        ),
    ] = None,
) -> None:
    """Tesseract: A toolkit for re-usable, autodiff-native software components."""
    verbose_tracebacks = loglevel == LogLevel.debug
    state.print_user_error_tracebacks = verbose_tracebacks
    app.pretty_exceptions_show_locals = verbose_tracebacks

    set_logger(loglevel, catch_warnings=True, rich_format=True)

    try:
        get_config()
    except PydanticValidationError as err:
        message = [
            "Error while parsing Tesseract configuration. "
            "Please check your environment variables.",
            "Errors found:",
        ]
        for error in err.errors():
            message.append(
                f' - TESSERACT_{str(error["loc"][0]).upper()}="{error["input"]}": {error["msg"]}'
            )
        raise UserError("\n".join(message)) from None


def _parse_config_override(
    options: list[str] | None,
) -> dict[tuple[str, ...], Any]:
    """Parse `["path1.path2.path3=value"]` into `[(["path1", "path2", "path3"], "value")]`."""
    if options is None:
        return {}

    def _parse_option(option: str) -> tuple[tuple[str, ...], Any]:
        if "=" not in option:
            raise typer.BadParameter(
                f'Invalid config override "{option}" (must be `keypath=value`)',
                param_hint="config_override",
            )

        key, value = option.split("=", maxsplit=1)
        if not re.match(r"\w[\w|\.]*", key):
            raise typer.BadParameter(
                f'Invalid keypath "{key}" in config override "{option}"',
                param_hint="config_override",
            )

        path = tuple(key.split("."))

        try:
            value = yaml.safe_load(value)
        except yaml.YAMLError as e:
            raise typer.BadParameter(
                f'Invalid value for config override "{option}", could not parse value as YAML: {e}',
                param_hint="config_override",
            ) from e

        return path, value

    return dict(_parse_option(option) for option in options)


@app.command("build")
@engine.needs_docker
def build_image(
    src_dir: Annotated[
        Path,
        typer.Argument(
            help=(
                "Source directory for the Tesseract. Must contain `tesseract_api.py` "
                "and `tesseract_config.yaml`."
            ),
            dir_okay=True,
            exists=True,
            file_okay=False,
            readable=True,
        ),
    ],
    tag: Annotated[
        str | None,
        typer.Option(
            "--tag",
            "-t",
            help="Tag for the resulting Tesseract. This can help you distinguish between "
            "Tesseracts with the same base name, like `cfd:v1` and cfd:v2`, "
            "both named `cfd` but tagged as `v1` and `v2` respectively.",
        ),
    ] = None,
    build_dir: Annotated[
        Path | None,
        typer.Option(
            help="Directory to use for the build. Defaults to a temporary directory.",
            dir_okay=True,
            file_okay=False,
            writable=True,
        ),
    ] = None,
    forward_ssh_agent: Annotated[
        bool,
        typer.Option(
            help=(
                "Forward the SSH agent to the Docker build environment. "
                "Has to be provided if requirements.txt contains private dependencies."
            ),
        ),
    ] = False,
    config_override: Annotated[
        list[str] | None,
        typer.Option(
            help=(
                "Override a configuration option in the Tesseract. "
                "Format: ``keypath=value`` where ``keypath`` is a dot-separated path to the "
                "attribute in tesseract_config.yaml. "
                "Possible keypaths are: "
                f"{', '.join(POSSIBLE_KEYPATHS)}. \n"
                "\n Possible build_config options are: "
                f"{', '.join(POSSIBLE_BUILD_CONFIGS)}. \n"
                "\nExample: ``--config-override build_config.target_platform=linux/arm64``."
            ),
            metavar="KEYPATH=VALUE",
        ),
    ] = None,
    generate_only: Annotated[
        bool,
        typer.Option(
            help="Only generate the build context and do not actually build the image."
        ),
    ] = False,
) -> None:
    """Build a new Tesseract from a context directory.

    The passed directory must contain the files `tesseract_api.py` and `tesseract_config.yaml`
    (can be created via `tesseract init`).

    Prints the built images as JSON array to stdout, for example: `["mytesseract:latest"]`.
    If `--generate-only` is set, the path to the build context is printed instead.
    """
    if config_override is None:
        config_override = []

    parsed_config_override = _parse_config_override(config_override)

    if generate_only:
        progress_indicator = nullcontext()
    else:
        progress_indicator = DEFAULT_CONSOLE.status(
            "[white]Processing", spinner="dots", spinner_style="white"
        )

    try:
        with progress_indicator:
            build_out = engine.build_tesseract(
                src_dir,
                tag,
                build_dir=build_dir,
                inject_ssh=forward_ssh_agent,
                config_override=parsed_config_override,
                generate_only=generate_only,
            )
    except BuildError as e:
        # raise from None to Avoid overly long tracebacks,
        # all the information is in the printed logs / exception str already
        raise UserError(f"Error building Tesseract: {e}") from None
    except APIError as e:
        raise UserError(f"Docker server error: {e}") from e
    except TypeError as e:
        raise UserError(f"Input error building Tesseract: {e}") from e
    except PermissionError as e:
        raise UserError(f"Permission denied: {e}") from e
    except ValidationError as e:
        raise UserError(f"Error validating tesseract_api.py: {e}") from e

    if generate_only:
        # output is the path to the build context
        build_dir = build_out
        typer.echo(build_dir)
    else:
        # output is the built image
        image = build_out
        logger.info(f"Built image {image.short_id}, {image.tags}")
        typer.echo(json.dumps(image.tags))


@app.command("init")
def init(
    name: Annotated[
        # Guaranteed to be a string by _validate_tesseract_name
        str | None,
        typer.Option(
            help="Tesseract name as specified in tesseract_config.yaml. Will be prompted if not provided.",
            callback=_validate_tesseract_name,
            show_default=False,
        ),
    ] = None,
    target_dir: Annotated[
        Path,
        typer.Option(
            help="Path to the directory where the Tesseract API module should be created.",
            dir_okay=True,
            file_okay=False,
            writable=True,
            show_default="current directory",
        ),
    ] = Path("."),
    recipe: Annotated[
        str,
        typer.Option(
            click_type=click.Choice(AVAILABLE_RECIPES),
            help="Use a pre-configured template to initialize Tesseract API and configuration.",
        ),
    ] = "base",
) -> None:
    """Initialize a new Tesseract API module."""
    logger.info(f"Initializing Tesseract {name} in directory: {target_dir}")
    engine.init_api(target_dir, name, recipe=recipe)


def _validate_port(port: str | None) -> str | None:
    """Validate port input."""
    if port is None:
        return None

    port = port.strip()
    if "-" in port:
        start, end = port.split("-")
    else:
        start = end = port

    try:
        start, end = int(start), int(end)
    except ValueError as ex:
        raise typer.BadParameter(
            (f"Port '{port}' must be single integer or a range (e.g. -p '8000-8080')."),
            param_hint="port",
        ) from ex

    if start > end:
        raise typer.BadParameter(
            (
                f"Start port '{start}' must be less than "
                f"or equal to end port '{end}' (e.g. -p '8000-8080')."
            ),
            param_hint="port",
        )

    if not (0 <= start <= 65535) or not (0 <= end <= 65535):
        raise typer.BadParameter(
            f"Ports '{port}' must be between 0 and 65535.",
            param_hint="port",
        )
    return port


@app.command("serve")
@engine.needs_docker
def serve(
    image_names: Annotated[
        list[str],
        typer.Argument(..., help="One or more Tesseract image names"),
    ],
    volume: Annotated[
        list[str] | None,
        typer.Option(
            "-v",
            "--volume",
            help="Bind mount a volume in all Tesseracts, in Docker format: source:target[:ro|rw]",
            metavar="source:target",
            show_default=False,
        ),
    ] = None,
    environment: Annotated[
        list[str] | None,
        typer.Option(
            "--env",
            "-e",
            help="Set environment variables in the Tesseract containers, in Docker format: key=value.",
        ),
    ] = None,
    port: Annotated[
        str | None,
        typer.Option(
            "--port",
            "-p",
            help="Optional port/port range to serve the Tesseract on (e.g. -p '8080-8082'). "
            "Port must be between 1025 and 65535.",
            callback=_validate_port,
        ),
    ] = None,
    host_ip: Annotated[
        str,
        typer.Option(
            "--host-ip",
            help=(
                "IP address of the host to bind the Tesseract to. "
                "Defaults to 127.0.0.1 (localhost). To bind to all interfaces, use '0.0.0.0'. "
                "WARNING: This may expose Tesseract to all local networks, use with caution."
            ),
        ),
    ] = "127.0.0.1",
    gpus: Annotated[
        list[str] | None,
        typer.Option(
            "--gpus",
            metavar="'all' | int",
            help=(
                "IDs of host Nvidia GPUs to make available in the Tesseract. "
                "You can use all GPUs via `--gpus all` or pass (multiple) IDs: `--gpus 0 --gpus 1`."
            ),
        ),
    ] = None,
    num_workers: Annotated[
        int,
        typer.Option(
            "--num-workers",
            help="Number of worker processes to use when serving the Tesseract.",
            show_default=True,
        ),
    ] = 1,
    debug: Annotated[
        bool,
        typer.Option(
            "--debug",
            help=(
                "Enable debug mode. This will propagate full tracebacks to the client "
                "and start a debugpy server in the Tesseract. "
                "WARNING: This may expose sensitive information, use with caution (and never in production)."
            ),
        ),
    ] = False,
    no_compose: Annotated[
        bool,
        typer.Option(
            "--no-compose",
            help=(
                "Do not use Docker Compose to serve the Tesseract. "
                "Instead, the command will block until interrupted. "
                "This is useful for cases in which docker-compose is not available."
            ),
        ),
    ] = False,
    service_names: Annotated[
        str | None,
        typer.Option(
            "--service-names",
            help=(
                "Comma-separated list of service names by which each Tesseract should be exposed "
                "in the shared network. "
                "Tesseracts are reachable from one another at http://{service_name}:8000. "
                "Not supported when using --no-compose."
            ),
        ),
    ] = None,
    user: Annotated[
        str | None,
        typer.Option(
            "--user",
            help=(
                "User to run the Tesseracts as e.g. '1000' or '1000:1000' (uid:gid). "
                "Defaults to the current user."
            ),
        ),
    ] = None,
    input_path: Annotated[
        str | None,
        typer.Option(
            "--input-path",
            help=(
                "Input path to read input files from, such as local directory or S3 URI "
                "(may be anything supported by fsspec)."
            ),
            hidden=True,
        ),
    ] = None,
) -> None:
    """Serve one or more Tesseract images.

    A successful serve command will display on standard output a JSON object
    with the Docker Compose project ID, which is required to run the teardown
    command, as well as a list of all containers spawned and their respective
    ports.
    """
    if port is not None:
        if len(image_names) > 1:
            # TODO: Docker compose with multiple ports is not supported until
            # docker/compose#7188 is resolved.
            raise typer.BadParameter(
                (
                    "Port specification only works if exactly one Tesseract is being served. "
                    f"Currently attempting to serve `{len(image_names)}` Tesseracts."
                ),
                param_hint="image_names",
            )
        ports = [port]
    else:
        ports = None

    # Parse environment variables from list to dict
    if environment is not None:
        try:
            environment = {
                env.split("=", maxsplit=1)[0]: env.split("=", maxsplit=1)[1]
                for env in environment
            }
        except Exception as ex:
            raise typer.BadParameter(
                "Environment variables must be in the format 'key=value'.",
                param_hint="environment",
            ) from ex

    if service_names is not None:
        if no_compose:
            raise typer.BadParameter(
                "Service name specification only works with Docker Compose.",
                param_hint="service_names",
            )

        service_names_list = service_names.split(",")
    else:
        service_names_list = None

    try:
        project_id = engine.serve(
            image_names,
            host_ip,
            ports,
            volume,
            environment,
            gpus,
            debug,
            num_workers,
            no_compose,
            service_names_list,
            user,
            input_path=input_path,
        )
    except RuntimeError as ex:
        raise UserError(
            f"Internal Docker error occurred while serving Tesseracts: {ex}"
        ) from ex

    container_ports = _display_project_meta(project_id)
    logger.info(
        f"Tesseract project ID, use it with 'tesseract teardown' command: {project_id}"
    )
    project_meta = {"project_id": project_id, "containers": container_ports}
    json_info = json.dumps(project_meta)
    typer.echo(json_info, nl=False)


@app.command("list")
@engine.needs_docker
def list_tesseract_images() -> None:
    """Display all Tesseract images."""
    _display_tesseract_image_meta()


@app.command("ps")
@engine.needs_docker
def list_tesseract_containers() -> None:
    """Display all Tesseract containers."""
    _display_tesseract_containers_meta()


def _display_tesseract_image_meta() -> None:
    """Display Tesseract image metadata."""
    table = RichTable("ID", "Tags", "Name", "Version", "Description")
    images = docker_client.images.list()
    for image in images:
        tesseract_vals = _get_tesseract_env_vals(image)
        if tesseract_vals:
            table.add_row(
                # Checksum Type + First 12 Chars of ID
                image.id[:19],
                str(image.tags),
                tesseract_vals["TESSERACT_NAME"],
                tesseract_vals.get("TESSERACT_VERSION", ""),
                tesseract_vals.get("TESSERACT_DESCRIPTION", "").replace("\n", " "),
            )
    RichConsole().print(table)


def _display_tesseract_containers_meta() -> None:
    """Display Tesseract containers metadata."""
    table = RichTable(
        "ID", "Name", "Version", "Host Address", "Project ID", "Description"
    )
    containers = docker_client.containers.list()
    for container in containers:
        tesseract_vals = _get_tesseract_env_vals(container)
        if tesseract_vals:
            tesseract_project = _find_tesseract_project(container)
            table.add_row(
                container.id[:12],
                tesseract_vals["TESSERACT_NAME"],
                tesseract_vals["TESSERACT_VERSION"],
                f"{container.host_ip}:{container.host_port}",
                tesseract_project,
                tesseract_vals.get("TESSERACT_DESCRIPTION", "").replace("\\n", " "),
            )
    RichConsole().print(table)


def _get_tesseract_env_vals(
    docker_asset: Image | Container,
) -> dict:
    """Convert Tesseract environment variables from list to dictionary."""
    env_vals = [s for s in docker_asset.attrs["Config"]["Env"] if "TESSERACT_" in s]
    return {item.split("=")[0]: item.split("=")[1] for item in env_vals}


def _find_tesseract_project(
    tesseract: Container,
) -> str:
    """Find the Tesseract Project ID for a given tesseract.

    A Tesseract Project ID is either a Docker Compose ID or a container name.
    """
    if tesseract.project_id is not None:
        return tesseract.project_id

    tesseract_id = tesseract.id[:12]

    for project, containers in docker_client.compose.list().items():
        if tesseract_id in containers:
            return project

    return tesseract.name


@app.command("apidoc")
@engine.needs_docker
def apidoc(
    image_name: Annotated[
        str,
        typer.Argument(..., help="Tesseract image name"),
    ],
    browser: Annotated[
        bool,
        typer.Option(help="Open the browser after serving the OpenAPI schema"),
    ] = True,
) -> None:
    """Serve the OpenAPI schema for a Tesseract."""
    host_ip = "127.0.0.1"
    project_id = engine.serve([image_name], no_compose=True, host_ip=host_ip)
    try:
        container = docker_client.containers.get(project_id)
        url = f"http://{host_ip}:{container.host_port}/docs"
        logger.info(f"Serving OpenAPI docs for Tesseract {image_name} at {url}")
        logger.info("  Press Ctrl+C to stop")
        if browser:
            webbrowser.open(url)
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            return
    finally:
        engine.teardown(project_id)


def _display_project_meta(project_id: str) -> list:
    """Display project metadata.

    Returns a list of dictionaries {name: container_name, port: host_port}.
    """
    container_ports = []

    compose_projects = docker_client.compose.list()
    if project_id in compose_projects:
        containers = compose_projects[project_id]
    else:
        containers = [project_id]

    for container_id in containers:
        container = docker_client.containers.get(container_id)
        logger.info(f"Container ID: {container.id}")
        logger.info(f"Name: {container.name}")
        entrypoint = container.attrs["Config"]["Entrypoint"]
        logger.info(f"Entrypoint: {entrypoint}")
        host_port = container.host_port
        host_ip = container.host_ip
        logger.info(f"View Tesseract: http://{host_ip}:{host_port}/docs")
        container_ports.append(
            {"name": container.name, "port": host_port, "ip": host_ip}
        )
        if container.host_debugpy_port:
            logger.info(
                f"Debugpy server listening at http://{host_ip}:{container.host_debugpy_port}"
            )

    return container_ports


@app.command("teardown")
@engine.needs_docker
def teardown(
    project_ids: Annotated[
        list[str] | None,
        typer.Argument(..., help="Docker Compose project IDs"),
    ] = None,
    tear_all: Annotated[
        bool,
        typer.Option(
            "--all",
            help="Tear down all Tesseracts currently being served.",
        ),
    ] = False,
) -> None:
    """Tear down one or more Tesseracts that were previously started with `tesseract serve`.

    One or more Tesseract project ids must be specified unless `--all` is set.
    """
    if not project_ids and not tear_all:
        raise typer.BadParameter(
            "Either project IDs or --all flag must be provided",
            param_hint="project_ids",
        )

    if project_ids and tear_all:
        raise typer.BadParameter(
            "Either project IDs or --all flag must be provided, but not both",
            param_hint="project_ids",
        )

    try:
        if not project_ids:
            project_ids = []  # Pass in empty list if no project_ids are provided.
        engine.teardown(project_ids, tear_all=tear_all)
    except ValueError as ex:
        raise UserError(
            f"Input error occurred while tearing down Tesseracts: {ex}"
        ) from ex
    except RuntimeError as ex:
        raise UserError(
            f"Internal Docker error occurred while tearing down Tesseracts: {ex}"
        ) from ex
    except NotFound as ex:
        raise UserError(f"Tesseract Project ID not found: {ex}") from ex


def _sanitize_error_output(error_output: str, tesseract_image: str) -> str:
    """Remove references to tesseract-runtime and unavailable commands from error output."""
    # Replace references to tesseract-runtime with tesseract run
    error_output = re.sub(
        r"Try 'tesseract-runtime",
        f"Try 'tesseract run {tesseract_image}",
        error_output,
    )

    error_output = re.sub(
        r"Usage: tesseract-runtime",
        f"Usage: tesseract run {tesseract_image}",
        error_output,
    )

    # Hide commands in help strings that users are not supposed to use via tesseract run
    error_output = re.sub(
        r"^│\s+(serve|health)\s+.*?$\n",
        "",
        error_output,
        flags=re.MULTILINE,
    )

    return error_output


@app.command(
    "run",
    # We need to ignore unknown options to forward the args to the Tesseract container
    context_settings={"ignore_unknown_options": True},
    # We implement --help manually to forward the help of the Tesseract container
    add_help_option=False,
)
@engine.needs_docker
def run_container(
    context: click.Context,
    tesseract_image: Annotated[
        str,
        typer.Argument(help="Tesseract image name"),
    ],
    cmd: Annotated[
        str,
        typer.Argument(help="Tesseract command to run"),
    ] = "",
    args: Annotated[
        list[str] | None,
        typer.Argument(help="Arguments for the command"),
    ] = None,
    volume: Annotated[
        list[str] | None,
        typer.Option(
            "-v",
            "--volume",
            help="Bind mount a volume, in Docker format: source:target.",
            metavar="source:target",
            show_default=False,
        ),
    ] = None,
    gpus: Annotated[
        list[str] | None,
        typer.Option(
            "--gpus",
            metavar="'all' | int",
            help=(
                "IDs of host GPUs to make available in the tesseract. "
                "You can use all GPUs via `--gpus all` or pass (multiple) IDs: `--gpus 0 --gpus 1`."
            ),
        ),
    ] = None,
    environment: Annotated[
        list[str] | None,
        typer.Option(
            "--env",
            "-e",
            help="Set environment variables in the Tesseract container, in Docker format: key=value.",
            metavar="key=value",
            show_default=False,
        ),
    ] = None,
    network: Annotated[
        str | None,
        typer.Option(
            "--network",
            help="Network to use for the Tesseract container, analogous to Docker's --network option. "
            "For example, 'host' uses the host system's network. Alternatively, you can create a custom "
            "network with `docker network create <network-name>` and use it here.",
            show_default=False,
        ),
    ] = None,
    user: Annotated[
        str | None,
        typer.Option(
            "--user",
            help=(
                "User to run the Tesseract as e.g. '1000' or '1000:1000' (uid:gid). "
                "Defaults to the current user."
            ),
        ),
    ] = None,
) -> None:
    """Execute a command in a Tesseract.

    This command starts a Tesseract instance and executes the given
    command.
    """
    if args is None:
        args = []

    if cmd == "serve":
        logger.error(
            "You should not serve tesseracts via "
            "`tesseract run <tesseract-name> serve`. "
            "Use `tesseract serve <tesseract-name>` instead."
        )
        raise typer.Exit(1)

    help_args = {"-h", "--help"}

    # When called as `tesseract run --help` -> show generic help
    if tesseract_image in help_args:
        context.get_help()
        return

    invoke_help = any(arg in help_args for arg in args) or cmd in help_args

    if (not cmd or cmd not in POSSIBLE_CMDS) and not invoke_help:
        if not cmd:
            error_string = f"Command is required. Are you sure your Tesseract image name is `{tesseract_image}`?\n"
        else:
            error_string = f"Command `{cmd}` does not exist. \n"

        error_string += (
            f"\nRun `tesseract run {tesseract_image} --help` for more information.\n"
        )

        error_string = (
            error_string + f"\nPossible commands are: {', '.join(POSSIBLE_CMDS)}"
        )
        raise typer.BadParameter(error_string, param_hint="cmd")

    if environment is not None:
        try:
            environment = {
                item.split("=", maxsplit=1)[0]: item.split("=", maxsplit=1)[1]
                for item in environment
            }
        except Exception as ex:
            raise typer.BadParameter(
                "Environment variables must be in the format 'key=value'.",
                param_hint="env",
            ) from ex
    try:
        result_out, result_err = engine.run_tesseract(
            tesseract_image,
            cmd,
            args,
            volumes=volume,
            gpus=gpus,
            environment=environment,
            network=network,
            user=user,
        )

    except ImageNotFound as e:
        raise UserError(
            "Tesseract image not found. "
            f"Are you sure your tesseract image name is {tesseract_image}?\n\n{e}"
        ) from e

    except ContainerError as e:
        msg = e.stderr.decode("utf-8").strip()
        if "No such command" in msg:
            error_string = f"Error running Tesseract '{tesseract_image}' \n\n Error: Unimplemented command '{cmd}'.  "
        else:
            error_string = _sanitize_error_output(
                f"Error running Tesseract. \n\n{msg}", tesseract_image
            )

        raise UserError(error_string) from e

    if invoke_help:
        result_err = _sanitize_error_output(result_err, tesseract_image)

    typer.echo(result_err, err=True, nl=False)
    typer.echo(result_out, nl=False)


def entrypoint() -> NoReturn:
    """Entrypoint for the Tesseract CLI."""
    try:
        result = app()
    except UserError as e:
        if state.print_user_error_tracebacks:
            # Do not print the exception here since it's part of the traceback
            logger.error("UserError occurred, traceback:", exc_info=True)
        else:
            # Prints only the error message without traceback
            logger.error(str(e), exc_info=False)
        result = 1
    except Exception:
        logger.critical("Uncaught error", exc_info=True)
        result = 2

    if result > 0:
        logger.critical("Aborting")

    raise SystemExit(result)


# Expose the underlying click object for doc generation
typer_click_object = typer.main.get_command(app)

if __name__ == "__main__":
    entrypoint()



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Any

import equinox as eqx
import jax
import jax.numpy as jnp
from pydantic import BaseModel, Field, model_validator
from typing_extensions import Self

from tesseract_core.runtime import Array, Differentiable, Float32
from tesseract_core.runtime.tree_transforms import filter_func, flatten_with_paths


class Vector_and_Scalar(BaseModel):
    v: Differentiable[Array[(None,), Float32]] = Field(
        description="An arbitrary vector"
    )
    s: Differentiable[Float32] = Field(description="A scalar", default=1.0)

    def scale(self) -> Differentiable[Array[(None,), Float32]]:
        return self.s * self.v


class InputSchema(BaseModel):
    a: Vector_and_Scalar = Field(
        description="An arbitrary vector and a scalar to multiply it by"
    )
    b: Vector_and_Scalar = Field(
        description="An arbitrary vector and a scalar to multiply it by "
        "must be of same shape as b"
    )
    norm_ord: int = Field(
        description="Order of norm (see numpy.linalg.norm)",
        default=2,
    )

    @model_validator(mode="after")
    def validate_shape_inputs(self) -> Self:
        if self.a.v.shape != self.b.v.shape:
            raise ValueError(
                f"a.v and b.v must have the same shape. "
                f"Got {self.a.v.shape} and {self.b.v.shape} instead."
            )
        return self


class Result_and_Norm(BaseModel):
    result: Differentiable[Array[(None,), Float32]] = Field(
        description="Vector s_a·a + s_b·b"
    )
    normed_result: Differentiable[Array[(None,), Float32]] = Field(
        description="Normalized Vector s_a·a + s_b·b/|s_a·a + s_b·b|"
    )


class OutputSchema(BaseModel):
    vector_add: Result_and_Norm
    vector_min: Result_and_Norm


@eqx.filter_jit
def apply_jit(inputs: dict) -> dict:
    a_scaled = inputs["a"]["s"] * inputs["a"]["v"]
    b_scaled = inputs["b"]["s"] * inputs["b"]["v"]
    add_result = a_scaled + b_scaled
    min_result = a_scaled - b_scaled
    return {
        "vector_add": {
            "result": add_result,
            "normed_result": add_result
            / jnp.linalg.norm(add_result, ord=inputs["norm_ord"]),
        },
        "vector_min": {
            "result": min_result,
            "normed_result": min_result
            / jnp.linalg.norm(min_result, ord=inputs["norm_ord"]),
        },
    }


def apply(inputs: InputSchema) -> OutputSchema:
    """Multiplies a vector `a` by `s`, and sums the result to `b`."""
    return apply_jit(inputs.model_dump())


def abstract_eval(abstract_inputs):
    """Calculate output shape of apply from the shape of its inputs."""
    is_shapedtye_dict = lambda x: type(x) is dict and (x.keys() == {"shape", "dtype"})
    is_shapedtye_struct = lambda x: isinstance(x, jax.ShapeDtypeStruct)

    jaxified_inputs = jax.tree.map(
        lambda x: jax.ShapeDtypeStruct(**x) if is_shapedtye_dict(x) else x,
        abstract_inputs.model_dump(),
        is_leaf=is_shapedtye_dict,
    )
    dynamic_inputs, static_inputs = eqx.partition(
        jaxified_inputs, filter_spec=is_shapedtye_struct
    )

    def wrapped_apply(dynamic_inputs):
        inputs = eqx.combine(static_inputs, dynamic_inputs)
        return apply_jit(inputs)

    jax_shapes = jax.eval_shape(wrapped_apply, dynamic_inputs)
    return jax.tree.map(
        lambda x: {"shape": x.shape, "dtype": str(x.dtype)}
        if is_shapedtye_struct(x)
        else x,
        jax_shapes,
        is_leaf=is_shapedtye_struct,
    )


def jacobian_vector_product(
    inputs: InputSchema,
    jvp_inputs: set[str],
    jvp_outputs: set[str],
    tangent_vector: dict[str, Any],
):
    return jvp_jit(
        inputs.model_dump(),
        tuple(jvp_inputs),
        tuple(jvp_outputs),
        tangent_vector,
    )


def vector_jacobian_product(
    inputs: InputSchema,
    vjp_inputs: set[str],
    vjp_outputs: set[str],
    cotangent_vector: dict[str, Any],
):
    return vjp_jit(
        inputs.model_dump(),
        tuple(vjp_inputs),
        tuple(vjp_outputs),
        cotangent_vector,
    )


def jacobian(
    inputs: InputSchema,
    jac_inputs: set[str],
    jac_outputs: set[str],
):
    return jac_jit(inputs.model_dump(), tuple(jac_inputs), tuple(jac_outputs))


@eqx.filter_jit
def jvp_jit(
    inputs: dict, jvp_inputs: tuple[str], jvp_outputs: tuple[str], tangent_vector: dict
):
    filtered_apply = filter_func(apply_jit, inputs, jvp_outputs)
    return jax.jvp(
        filtered_apply,
        [flatten_with_paths(inputs, include_paths=jvp_inputs)],
        [tangent_vector],
    )[1]


@eqx.filter_jit
def vjp_jit(
    inputs: dict,
    vjp_inputs: tuple[str],
    vjp_outputs: tuple[str],
    cotangent_vector: dict,
):
    filtered_apply = filter_func(apply_jit, inputs, vjp_outputs)
    _, vjp_func = jax.vjp(
        filtered_apply, flatten_with_paths(inputs, include_paths=vjp_inputs)
    )
    return vjp_func(cotangent_vector)[0]


@eqx.filter_jit
def jac_jit(
    inputs: dict,
    jac_inputs: tuple[str],
    jac_outputs: tuple[str],
):
    filtered_apply = filter_func(apply_jit, inputs, jac_outputs)
    return jax.jacrev(filtered_apply)(
        flatten_with_paths(inputs, include_paths=jac_inputs)
    )



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import json
import traceback

import docker.errors
from typer.testing import CliRunner

from tesseract_core.sdk.cli import app
from tesseract_core.sdk.docker_client import (
    CLIDockerClient,
    ImageNotFound,
    NotFound,
)


def image_exists(client, image_name, tesseract_only: bool = True):
    """Checks if image name exists."""
    # Docker images may be prefixed with the registry URL
    kwargs = {}
    if isinstance(client, CLIDockerClient):
        kwargs["tesseract_only"] = tesseract_only

    try:
        client.images.get(image_name, **kwargs)
        return True
    except (ImageNotFound, docker.errors.ImageNotFound):
        return False


def container_exists(client, container_name_or_id, tesseract_only: bool = True):
    """Checks if containers exists."""
    try:
        client.containers.get(container_name_or_id, tesseract_only)
        return True
    except NotFound:
        return False


def print_debug_info(result):
    """Print debug info from result of a CLI command if it failed."""
    if result.exit_code == 0:
        return
    print(result.output)
    if result.exc_info:
        traceback.print_exception(*result.exc_info)


def build_tesseract(
    client, sourcedir, image_name, config_override=None, tag=None, build_retries=3
):
    cli_runner = CliRunner(mix_stderr=False)

    build_args = [
        "--loglevel",
        "debug",
        "build",
        str(sourcedir),
        "--config-override",
        f"name={image_name}",
    ]

    if config_override is not None:
        for key, val in config_override.items():
            build_args.extend(["--config-override", f"{key}={val}"])

    if tag is not None:
        build_args.extend(["--tag", tag])
        image_name = f"{image_name}:{tag}"
    else:
        image_name = f"{image_name}:latest"

    for _ in range(build_retries):
        result = cli_runner.invoke(
            app,
            build_args,
            catch_exceptions=False,
        )
        # Retry if the build fails with EOF error (connectivity issue)
        # See https://github.com/docker/buildx/issues/2064
        is_expected_err = "error reading from server: EOF" in result.output
        if not is_expected_err:
            break

    print_debug_info(result)
    assert result.exit_code == 0, result.exception

    image_tag = json.loads(result.stdout.strip())[0]

    # This raise an error if the image does not exist
    client.images.get(image_tag)
    return image_tag



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import ast
from pathlib import Path
from typing import Annotated, Literal, NamedTuple, Union

import yaml
from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    Strict,
)
from pydantic import ValidationError as PydanticValidationError


class _ApiObject(NamedTuple):
    name: str
    expected_type: type
    num_args: int | None = None
    arg_names: tuple[str, ...] | None = None
    optional: bool = False


ORDINALS = ["first", "second", "third", "fourth", "fifth", "sixth", "seventh", "eighth"]

EXPECTED_OBJECTS = (
    _ApiObject("apply", ast.FunctionDef, 1, arg_names=("inputs",)),
    _ApiObject("InputSchema", ast.ClassDef),
    _ApiObject("OutputSchema", ast.ClassDef),
    _ApiObject(
        "jacobian",
        ast.FunctionDef,
        3,
        arg_names=("inputs", "jac_inputs", "jac_outputs"),
        optional=True,
    ),
    _ApiObject(
        "jacobian_vector_product",
        ast.FunctionDef,
        4,
        arg_names=("inputs", "jvp_inputs", "jvp_outputs", "tangent_vector"),
        optional=True,
    ),
    _ApiObject(
        "vector_jacobian_product",
        ast.FunctionDef,
        4,
        arg_names=("inputs", "vjp_inputs", "vjp_outputs", "cotangent_vector"),
        optional=True,
    ),
    _ApiObject(
        "abstract_eval",
        ast.FunctionDef,
        1,
        arg_names=("abstract_inputs",),
        optional=True,
    ),
)


def assert_relative_path(value: str) -> str:
    """Assert that a string encodes a relative path."""
    if Path(value).is_absolute():
        raise ValueError(f"value must be a relative path (got {value})")
    return value


RelativePath = Annotated[str, AfterValidator(assert_relative_path)]
StrictStr = Annotated[str, Strict()]


class PipRequirements(BaseModel):
    """Configuration options for Python environments built via pip."""

    provider: Literal["python-pip"]
    _filename: Literal["tesseract_requirements.txt"] = "tesseract_requirements.txt"
    _build_script: Literal["build_pip_venv.sh"] = "build_pip_venv.sh"
    model_config: ConfigDict = ConfigDict(extra="forbid")


class CondaRequirements(BaseModel):
    """Configuration options for Python environments built via conda."""

    provider: Literal["conda"]
    _filename: Literal["tesseract_environment.yaml"] = "tesseract_environment.yaml"
    _build_script: Literal["build_conda_venv.sh"] = "build_conda_venv.sh"
    model_config: ConfigDict = ConfigDict(extra="forbid")


PythonRequirements = Union[PipRequirements, CondaRequirements]


class TesseractBuildConfig(BaseModel, validate_assignment=True):
    """Configuration options for building a Tesseract."""

    base_image: StrictStr = Field(
        "debian:bookworm-slim",
        description="Base Docker image for the build. Must be Debian-based.",
    )
    target_platform: StrictStr = Field(
        "native",
        description=(
            "Target platform for the Docker build. Must be a valid Docker platform, "
            "or 'native' to build for the host platform. "
            "In general, images built for one platform will not run on another."
        ),
    )
    extra_packages: tuple[StrictStr, ...] = Field(
        (), description="Extra packages to install during build via apt-get."
    )
    package_data: tuple[tuple[RelativePath, StrictStr], ...] | None = Field(
        (),
        description=(
            "Additional files to copy into the Docker image, in the format ``(source, destination)``. "
            "Source paths are relative to the Tesseract directory."
        ),
    )
    custom_build_steps: tuple[StrictStr, ...] | None = Field(
        (),
        description=(
            "Custom steps to run during ``docker build`` (after everything else is installed). "
            "Example: ``[\"RUN echo 'Hello, world!'\"]``"
        ),
    )

    requirements: PythonRequirements = PipRequirements(provider="python-pip")

    model_config = ConfigDict(extra="forbid")

    skip_checks: bool = Field(
        False,
        description=(
            "If True, skip build-time checks of Tesseract API module. "
            "This can be useful when such a check cannot succeed (e.g. when building for a "
            "different platform), but may lead to runtime errors if the Tesseract API is not valid."
        ),
    )


# Allow None to be passed as a valid value for build_config, for example in YAML that comments out all options.
OptionalBuildConfig = Annotated[
    TesseractBuildConfig,
    BeforeValidator(lambda v: TesseractBuildConfig() if v is None else v),
]


class TesseractConfig(BaseModel, validate_assignment=True):
    """Configuration options for Tesseracts. Defines valid options in ``tesseract_config.yaml``."""

    name: StrictStr = Field(..., description="Name of the Tesseract.")
    version: StrictStr = Field("0+unknown", description="Version of the Tesseract.")
    description: StrictStr = Field(
        "",
        description="Free-text description of what the Tesseract does.",
    )
    build_config: OptionalBuildConfig = Field(
        default_factory=TesseractBuildConfig,
        description="Configuration options for building the Tesseract.",
    )

    model_config = ConfigDict(extra="forbid")


class ValidationError(Exception):
    """Raised when inputs needed to build a tesseract are invalid."""

    pass


def _get_func_argnames(func: ast.FunctionDef) -> tuple[str, ...]:
    """Get the names of the arguments of a function node.

    See:
    https://docs.python.org/3/library/ast.html#ast.FunctionDef
    https://docs.python.org/3/library/ast.html#ast.arguments
    """
    func_args = func.args
    if func_args.kwonlyargs:
        raise ValidationError(
            f"Function {func.name} must not have keyword-only arguments"
        )
    if func_args.posonlyargs:
        raise ValidationError(
            f"Function {func.name} must not have positional-only arguments"
        )
    return tuple(arg.arg for arg in func_args.args)


def validate_tesseract_api(src_dir: Path) -> None:
    """Check that given folder contains a Tesseract API that satisfies our constraints.

    This function does not return anything, but it raises `ValidationError` in
    case something goes wrong. In particular, we are checking that:
      *  The mandatory endpoints needed for a tesseract are actually
         implemented
      *  The implemented functions have the correct signature
      *  Both InputSchema and OutputSchema are `pydantic.BaseModel`s.

    Args:
        src_dir (Path): Path to the directory containing tesseract_api.py and tesseract_config.yaml.
    """
    tesseract_api_location = src_dir / "tesseract_api.py"
    config_location = src_dir / "tesseract_config.yaml"

    if not tesseract_api_location.exists():
        raise ValidationError(f"No file found at {tesseract_api_location}")

    if not config_location.exists():
        raise ValidationError(f"No file found at {config_location}")

    # Validate config
    try:
        get_config(src_dir)
    except PydanticValidationError as err:
        raise ValidationError(
            f"Invalid configuration in {config_location}: {err}"
        ) from err

    # Parse Tesseract API
    with open(tesseract_api_location) as f:
        tesseract_api_code = f.read()

    try:
        tesseract_api = ast.parse(tesseract_api_code)
    except SyntaxError as err:
        raise ValidationError(
            f"Syntax error in {tesseract_api_location}: {err}"
        ) from err

    # Check if expected attributes are defined
    toplevel_objects = {
        node.name: node for node in tesseract_api.body if hasattr(node, "name")
    }

    for obj in EXPECTED_OBJECTS:
        if obj.name not in toplevel_objects:
            if obj.optional:
                continue

            raise ValidationError(f"{obj.name} not defined in {tesseract_api_location}")

        if not isinstance(toplevel_objects[obj.name], obj.expected_type):
            raise ValidationError(
                f"{obj.name} is not a {obj.expected_type.__name__} in {tesseract_api_location}"
            )

        if obj.num_args is not None:
            func_argnames = _get_func_argnames(toplevel_objects[obj.name])
            func_argnums = len(func_argnames)
            if func_argnums != obj.num_args:
                raise ValidationError(
                    f"{obj.name} must have {obj.num_args} arguments: {', '.join(obj.arg_names)}.\n"
                    f"However, {tesseract_api_location} specifies {func_argnums} "
                    f"arguments: {', '.join(func_argnames)}."
                )
            msgs = []
            for i in range(obj.num_args):
                if func_argnames[i] != obj.arg_names[i]:
                    msgs.append(
                        f"The {ORDINALS[i]} argument (argument {i}) of {obj.name} must be named {obj.arg_names[i]}, "
                        f"but {tesseract_api_location} has named it {func_argnames[i]}."
                    )
            if msgs:
                raise ValidationError("\n".join(msgs))

    # Check InputSchema and OutputSchema are pydantic BaseModels
    for schema in ("InputSchema", "OutputSchema"):
        obj = toplevel_objects[schema]
        if not obj.bases:
            raise ValidationError(f"{schema} must inherit from pydantic.BaseModel")


def get_config(src_dir: Path) -> TesseractConfig:
    """Get configuration options from a tesseract_config.yaml file."""
    config_file = src_dir / "tesseract_config.yaml"

    if not config_file.exists():
        raise FileNotFoundError(f"No file found at {config_file}")

    with open(config_file) as f:
        config = yaml.safe_load(f)

    try:
        return TesseractConfig(**config)
    except PydanticValidationError as err:
        raise ValidationError(f"Invalid configuration: {err}") from err


def get_non_base_fields_in_tesseract_config() -> list[tuple[str, type]]:
    """Gets fields in Tesseract Config that are not a base fields."""
    base_fields = (str, int, float, bool, bytes)
    non_base_fields = []
    for field_name, field_info in TesseractConfig.model_fields.items():
        if field_info.annotation not in base_fields:
            non_base_fields.append((field_name, field_info.annotation))
    return non_base_fields



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from ._version import __version__ as scm_version

__version__ = scm_version

# import public API of the package
# from . import <obj>
from .sdk.engine import build_tesseract, run_tesseract, serve, teardown
from .sdk.tesseract import Tesseract

# add public API as strings here, for example __all__ = ["obj"]
__all__ = ["Tesseract", "build_tesseract", "run_tesseract", "serve", "teardown"]



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import inspect
from functools import wraps
from types import ModuleType
from typing import Any, Callable, Union

import uvicorn
from fastapi import FastAPI, Header, Response
from pydantic import BaseModel

from .config import get_config
from .core import create_endpoints
from .file_interactions import SUPPORTED_FORMATS, output_to_bytes

# Endpoints that should use GET instead of POST
GET_ENDPOINTS = {"input_schema", "output_schema", "health"}

# TODO: make this configurable via environment variable
DEFAULT_ACCEPT = "application/json"


def create_response(model: BaseModel, accept: str) -> Response:
    """Create a response of the format specified by the Accept header."""
    if accept is None or accept == "*/*":
        accept = DEFAULT_ACCEPT

    output_format: SUPPORTED_FORMATS = accept.split("/")[-1]
    content = output_to_bytes(model, output_format)

    return Response(status_code=200, content=content, media_type=accept)


def create_rest_api(api_module: ModuleType) -> FastAPI:
    """Create the Tesseract REST API."""
    config = get_config()
    app = FastAPI(
        title=config.name,
        version=config.version,
        description=config.description.replace("\\n", "\n"),
        docs_url=None,
        redoc_url="/docs",
        debug=config.debug,
    )
    tesseract_endpoints = create_endpoints(api_module)

    def wrap_endpoint(endpoint_func: Callable):
        endpoints_to_wrap = [
            "apply",
            "jacobian",
            "jacobian_vector_product",
            "vector_jacobian_product",
        ]

        @wraps(endpoint_func)
        async def wrapper(*args: Any, accept: str, **kwargs: Any):
            result = endpoint_func(*args, **kwargs)
            return create_response(result, accept)

        if endpoint_func.__name__ not in endpoints_to_wrap:
            return endpoint_func
        else:
            # wrapper's signarure will be the same as endpoint
            # func's signature. We do however need to change this
            # in order to add a Header parameter that FastAPI
            # will understand.
            original_sig = inspect.signature(endpoint_func)
            accept = inspect.Parameter(
                "accept",
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=Header(default=None),
                annotation=Union[str, None],
            )
            # Other header parameters common to computational endpoints
            # could be defined and appended here as well.
            new_params = [*list(original_sig.parameters.values()), accept]
            new_sig = original_sig.replace(parameters=new_params)
            wrapper.__signature__ = new_sig
            return wrapper

    for endpoint_func in tesseract_endpoints:
        endpoint_name = endpoint_func.__name__
        wrapped_endpoint = wrap_endpoint(endpoint_func)
        http_methods = ["GET"] if endpoint_name in GET_ENDPOINTS else ["POST"]
        app.add_api_route(f"/{endpoint_name}", wrapped_endpoint, methods=http_methods)

    return app


def serve(host: str, port: int, num_workers: int) -> None:
    """Start the REST API."""
    config = get_config()
    if config.debug:
        import debugpy

        debugpy.listen(("0.0.0.0", 5678))

    uvicorn.run(
        "tesseract_core.runtime.app_http:app", host=host, port=port, workers=num_workers
    )



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Any

import numpy as np
import torch
from pydantic import BaseModel, Field, model_validator
from torch.utils._pytree import tree_map
from typing_extensions import Self

from tesseract_core.runtime import Array, Differentiable, Float32
from tesseract_core.runtime.tree_transforms import filter_func, flatten_with_paths

#
# Schemata
#


class Vector_and_Scalar(BaseModel):
    v: Differentiable[Array[(None,), Float32]] = Field(
        description="An arbitrary vector"
    )
    s: Differentiable[Float32] = Field(description="A scalar", default=1.0)

    # we lose the ability to use methods such as this when using model_dump
    # unless we reconstruct nested models
    def scale(self) -> Differentiable[Array[(None,), Float32]]:
        return self.s * self.v


class InputSchema(BaseModel):
    a: Vector_and_Scalar = Field(
        description="An arbitrary vector and a scalar to multiply it by"
    )
    b: Vector_and_Scalar = Field(
        description="An arbitrary vector and a scalar to multiply it by "
        "must be of same shape as b"
    )

    @model_validator(mode="after")
    def validate_shape_inputs(self) -> Self:
        if self.a.v.shape != self.b.v.shape:
            raise ValueError(
                f"a.v and b.v must have the same shape. "
                f"Got {self.a.v.shape} and {self.b.v.shape} instead."
            )
        return self


class Result_and_Norm(BaseModel):
    result: Differentiable[Array[(None,), Float32]] = Field(
        description="Vector s_a·a + s_b·b"
    )
    normed_result: Differentiable[Array[(None,), Float32]] = Field(
        description="Normalized Vector s_a·a + s_b·b/|s_a·a + s_b·b|"
    )


class OutputSchema(BaseModel):
    vector_add: Result_and_Norm
    vector_min: Result_and_Norm


#
# Required endpoints
#


def evaluate(inputs: Any) -> Any:
    a_scaled = inputs["a"]["s"] * inputs["a"]["v"]
    b_scaled = inputs["b"]["s"] * inputs["b"]["v"]
    add_result = a_scaled + b_scaled
    min_result = a_scaled - b_scaled
    return {
        "vector_add": {
            "result": add_result,
            "normed_result": add_result / torch.linalg.norm(add_result, ord=2),
        },
        "vector_min": {
            "result": min_result,
            "normed_result": min_result / torch.linalg.norm(min_result, ord=2),
        },
    }


def apply(inputs: InputSchema) -> OutputSchema:
    # Optional: Insert any pre-processing/setup that doesn't require tracing
    # and is only required when specifically running your apply function
    # and not your differentiable endpoints.
    # For example, you might want to set up a logger or mlflow server.
    # Pre-processing should not modify any input that could impact the
    # differentiable outputs in a nonlinear way (a constant shift
    # should be safe)

    # Convert to pytorch tensors
    tensor_inputs = tree_map(to_tensor, inputs.model_dump())
    out = evaluate(tensor_inputs)

    # Optional: Insert any post-processing that doesn't require tracing
    # For example, you might want to save to disk or modify a non-differentiable
    # output. Again, do not modify any differentiable output in a non-linear way.
    return out


#
# Pytorch-handled AD endpoints (no need to modify)
#


def jacobian(
    inputs: InputSchema,
    jac_inputs: set[str],
    jac_outputs: set[str],
):
    # convert all numbers and arrays to torch tensors
    tensor_inputs = tree_map(to_tensor, inputs.model_dump())

    # flatten the dict such that they can be accessed by paths and obtain the values
    pos_inputs = flatten_with_paths(tensor_inputs, jac_inputs).values()

    # create a positional function that accepts a flat sequence of values and returns a tuple
    filtered_pos_eval = filter_func(
        evaluate,
        tensor_inputs,
        jac_outputs,
        input_paths=jac_inputs,
    )

    def filtered_pos_eval_flat(*args):
        res = filtered_pos_eval(*args)
        return tuple(res[k] for k in jac_outputs)

    # calculate the jacobian
    jacobian = torch.autograd.functional.jacobian(
        filtered_pos_eval_flat, tuple(pos_inputs)
    )

    # rebuild the dictionary from the list of results
    jac_dict = {}
    for dy, dys in zip(jac_outputs, jacobian, strict=True):
        jac_dict[dy] = {}
        for dx, dxs in zip(jac_inputs, dys, strict=True):
            jac_dict[dy][dx] = dxs

    return jac_dict


def jacobian_vector_product(
    inputs: InputSchema,
    jvp_inputs: set[str],
    jvp_outputs: set[str],
    tangent_vector: dict[str, Any],
):
    # Make ordering of tangent_vector identical to jvp_inputs
    tangent_vector = {key: tangent_vector[key] for key in jvp_inputs}

    # convert all numbers and arrays to torch tensors
    tensor_inputs = tree_map(to_tensor, inputs.model_dump())
    pos_tangent = tree_map(to_tensor, tangent_vector).values()

    # flatten the dictionaries such that they can be accessed by paths
    pos_inputs = flatten_with_paths(tensor_inputs, jvp_inputs).values()

    # create a positional function that accepts a list of values
    filtered_pos_eval = filter_func(
        evaluate, tensor_inputs, jvp_outputs, input_paths=jvp_inputs
    )

    return torch.func.jvp(filtered_pos_eval, tuple(pos_inputs), tuple(pos_tangent))[1]


def vector_jacobian_product(
    inputs: InputSchema,
    vjp_inputs: set[str],
    vjp_outputs: set[str],
    cotangent_vector: dict[str, Any],
):
    # Make ordering of cotangent_vector identical to vjp_inputs
    cotangent_vector = {key: cotangent_vector[key] for key in vjp_outputs}

    # convert all numbers and arrays to torch tensors
    tensor_inputs = tree_map(to_tensor, inputs.model_dump())
    tensor_cotangent = tree_map(to_tensor, cotangent_vector)

    # flatten the dictionaries such that they can be accessed by paths
    pos_inputs = flatten_with_paths(tensor_inputs, vjp_inputs).values()

    # create a positional function that accepts a list of values
    filtered_pos_func = filter_func(
        evaluate, tensor_inputs, vjp_outputs, input_paths=vjp_inputs
    )

    _, vjp_func = torch.func.vjp(filtered_pos_func, *pos_inputs)

    vjp_vals = vjp_func(tensor_cotangent)
    return dict(zip(vjp_inputs, vjp_vals, strict=True))


to_tensor = lambda x: torch.tensor(x) if isinstance(x, np.generic | np.ndarray) else x



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Tests for the MPA library."""

import csv
import json
import os

import pytest

from tesseract_core.runtime import mpa
from tesseract_core.runtime.experimental import (
    log_artifact,
    log_metric,
    log_parameter,
    start_run,
)


def test_start_run_context_manager():
    """Test that start_run works as a context manager."""
    with start_run():
        log_parameter("test_param", "value")
        log_metric("test_metric", 0.5)


def test_no_active_run_error():
    """Test that logging functions raise error when no run is active."""
    with pytest.raises(RuntimeError, match="No active MPA run"):
        log_parameter("test", "value")

    with pytest.raises(RuntimeError, match="No active MPA run"):
        log_metric("test", 0.5)

    with pytest.raises(RuntimeError, match="No active MPA run"):
        log_artifact("test.txt")


def test_nested_runs():
    """Test that nested runs work correctly."""
    with start_run():
        log_parameter("outer", "value1")

        with start_run():
            log_parameter("inner", "value2")
            log_metric("inner_metric", 1.0)

        # Should still work in outer context
        log_parameter("outer2", "value3")


def test_file_backend_default():
    """Test that FileBackend is used by default."""
    backend = mpa._create_backend()
    assert isinstance(backend, mpa.FileBackend)


def test_file_backend_empty_mlflow_uri():
    """Test that FileBackend is used when MLFLOW_TRACKING_URI is empty."""
    os.environ["MLFLOW_TRACKING_URI"] = ""
    backend = mpa._create_backend()
    assert isinstance(backend, mpa.FileBackend)


def test_custom_log_directory(tmpdir):
    """Test that FileBackend respects MPA_DIR environment variable."""
    custom_dir = tmpdir / "custom_logs"
    os.environ["MPA_DIR"] = str(custom_dir)

    backend = mpa.FileBackend()
    assert backend.log_dir == custom_dir
    assert backend.log_dir.exists()


def test_unique_run_directories():
    """Test that each FileBackend instance creates unique run directories."""
    backend1 = mpa.FileBackend()
    backend2 = mpa.FileBackend()

    assert backend1.run_dir != backend2.run_dir
    assert backend1.run_dir.exists()
    assert backend2.run_dir.exists()


def test_log_parameter_content():
    """Test parameter logging creates correct JSON content."""
    backend = mpa.FileBackend()
    backend.log_parameter("model_name", "test_model")
    backend.log_parameter("epochs", 10)
    backend.log_parameter("learning_rate", 0.001)

    # Verify JSON file content
    assert backend.params_file.exists()
    with open(backend.params_file) as f:
        params = json.load(f)

    assert params["model_name"] == "test_model"
    assert params["epochs"] == 10
    assert params["learning_rate"] == 0.001


def test_log_metric_content():
    """Test metric logging creates correct CSV content."""
    backend = mpa.FileBackend()
    backend.log_metric("accuracy", 0.95)
    backend.log_metric("loss", 0.05, step=1)

    # Verify CSV file content
    assert backend.metrics_file.exists()
    with open(backend.metrics_file) as f:
        reader = csv.DictReader(f)
        metrics = list(reader)

    assert len(metrics) == 2

    # First metric (auto-generated step)
    assert metrics[0]["key"] == "accuracy"
    assert float(metrics[0]["value"]) == 0.95
    assert int(metrics[0]["step"]) == 0
    assert "timestamp" in metrics[0]

    # Second metric (explicit step)
    assert metrics[1]["key"] == "loss"
    assert float(metrics[1]["value"]) == 0.05
    assert int(metrics[1]["step"]) == 1


def test_log_artifact_content(tmpdir):
    """Test artifact logging copies files correctly."""
    backend = mpa.FileBackend()

    # Create a test file with specific content
    test_file = tmpdir / "model_summary.txt"
    test_content = "Model: CNN\nAccuracy: 95.2%\nLoss: 0.048"
    test_file.write_text(test_content, encoding="utf-8")

    # Log the artifact
    backend.log_artifact(str(test_file))

    # Verify file was copied with correct content
    copied_file = backend.artifacts_dir / "model_summary.txt"
    assert copied_file.exists()
    assert copied_file.read_text() == test_content


def test_log_artifact_missing_file():
    """Test that logging non-existent artifact raises error."""
    backend = mpa.FileBackend()

    with pytest.raises(FileNotFoundError, match="Artifact file not found"):
        backend.log_artifact("non_existent_file.txt")


def test_mlflow_backend_creation(tmpdir):
    """Test that MLflowBackend is created when MLFLOW_TRACKING_URI is set."""
    mlflow_dir = tmpdir / "mlflow_backend_test"
    os.environ["MLFLOW_TRACKING_URI"] = f"file://{mlflow_dir}"
    backend = mpa._create_backend()
    assert isinstance(backend, mpa.MLflowBackend)


def test_mlflow_log_calls(tmpdir):
    """Test MLflow backend logging functions with temporary directory."""
    mlflow_dir = tmpdir / "mlflow_logging_test"
    os.environ["MLFLOW_TRACKING_URI"] = f"file://{mlflow_dir}"

    with start_run():
        log_parameter("model_type", "neural_network")
        log_parameter("epochs", 100)

        log_metric("accuracy", 0.85)
        log_metric("loss", 0.25, step=1)

        artifact_file = tmpdir / "model_config.json"
        artifact_file.write_text("Test content", encoding="utf-8")
        log_artifact(str(artifact_file))

    # Verify MLflow directory structure was created
    assert mlflow_dir.exists()
    # MLflow creates experiment directories, so we should see some structure
    mlflow_contents = list(mlflow_dir.listdir())
    assert len(mlflow_contents) > 0



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from pydantic import BaseModel

from tesseract_core.runtime.experimental import log_artifact, log_metric, log_parameter


class InputSchema(BaseModel):
    pass


class OutputSchema(BaseModel):
    pass


def apply(inputs: InputSchema) -> OutputSchema:
    """This demonstrates logging parameters, metrics and artifacts."""
    log_parameter("example_param", "value")

    for step in range(10):
        metric_value = step**2
        log_metric("squared_step", metric_value, step=step)

    text = "This is an output file we want to log as an artifact."
    with open("/tmp/artifact.txt", "w") as f:
        f.write(text)

    log_artifact("/tmp/artifact.txt")
    return OutputSchema()



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import json
from collections.abc import Iterable
from copy import deepcopy
from typing import Annotated, Optional

import numpy as np
import pytest
from pydantic import BaseModel, ConfigDict, ValidationError

from tesseract_core.runtime import Array, Differentiable, Float32, Float64, Int64, UInt8
from tesseract_core.runtime.experimental import LazySequence
from tesseract_core.runtime.schema_generation import (
    apply_function_to_model_tree,
    create_abstract_eval_schema,
    create_apply_schema,
    create_autodiff_schema,
)


class SubModel(BaseModel):
    foo: Float32
    bar: list[Differentiable[Array[..., Int64]]]


class NestedModel(BaseModel):
    testdiffarr: Differentiable[Array[(5, None), Float64]]
    testfoo: list[SubModel] | None
    testbar: dict[str, SubModel]
    testbaz: Array[(1, 2, 3), UInt8]
    testset: set[int]
    testtuple: tuple[int, str]
    testlazysequence: LazySequence[tuple[str, Differentiable[Array[(None,), Float32]]]]


def make_array(shape, dtype):
    rng = np.random.default_rng()
    return rng.integers(0, 10, shape).astype(dtype)


def replace_arrays(inpobj, replace_fn):
    """Recursively replace NumPy arrays in nested inpobj by calling replace_fn(arr)."""
    if isinstance(inpobj, dict) and "shape" in inpobj and "dtype" in inpobj:
        return replace_fn(inpobj)
    elif isinstance(inpobj, dict):
        return {k: replace_arrays(v, replace_fn) for k, v in inpobj.items()}
    elif isinstance(inpobj, Iterable) and not isinstance(inpobj, str):
        return type(inpobj)([replace_arrays(v, replace_fn) for v in inpobj])
    else:
        return inpobj


testinput = NestedModel(
    testdiffarr=make_array((5, 6), "float64"),
    testfoo=[SubModel(foo=1.0, bar=[1, 2, 3]), SubModel(foo=4.0, bar=[])],
    testbar={"hey there": SubModel(foo=5.0, bar=[2])},
    testbaz=make_array((1, 2, 3), "uint8"),
    testset={1, 2, 3},
    testtuple=(1, "hello"),
    testlazysequence=[
        ("a", make_array((3,), "float32")),
        ("b", make_array((4,), "float32")),
    ],
).model_dump(
    mode="json",
    # This encoding simplifies how replace_arrays works.
    context={"array_encoding": "base64"},
)

testinput_arrays_only = {
    "testdiffarr": testinput["testdiffarr"],
    "testfoo": [
        {"bar": [testinput["testfoo"][0]["bar"][0]]},
        {"bar": []},
    ],
    "testbar": {"hey there": {"bar": [testinput["testbar"]["hey there"]["bar"][0]]}},
    "testlazysequence": [
        (testinput["testlazysequence"][0][1],),
        (testinput["testlazysequence"][1][1],),
    ],
}


def test_create_apply_schema():
    InputSchema, OutputSchema = create_apply_schema(NestedModel, NestedModel)

    InputSchema.model_validate({"inputs": testinput})
    OutputSchema.model_validate(testinput)

    # Extra keys
    with pytest.raises(ValidationError):
        InputSchema.model_validate({"inputs": testinput, "foo": 1})

    # Extra keys (nested dict)
    with pytest.raises(ValidationError):
        inp = testinput.copy()
        inp["foo"] = 1
        InputSchema.model_validate({"inputs": inp})

    # Extra keys (nested model)
    inp = deepcopy(testinput)
    inp["testfoo"][0] = {"foo": 1, "bar": [1, 2, 3]}
    InputSchema.model_validate({"inputs": inp})

    with pytest.raises(ValidationError):
        inp["testfoo"][0] = {"foo": 1, "bar": [1, 2, 3], "extra": 1}
        InputSchema.model_validate({"inputs": inp})

    # Extra keys (output)
    with pytest.raises(ValidationError):
        OutputSchema.model_validate({**testinput, "foo": 1})


def test_create_abstract_eval_schema():
    testinput_abstract = replace_arrays(
        testinput, lambda arr: {"shape": arr["shape"], "dtype": arr["dtype"]}
    )

    AbstractInputSchema, AbstractOutputSchema = create_abstract_eval_schema(
        NestedModel, NestedModel
    )

    UnwrappedInputSchema = AbstractInputSchema.model_fields["inputs"].annotation
    UnwrappedOutputSchema = AbstractOutputSchema.model_fields["root"].annotation

    # should have identical structure
    assert NestedModel.model_fields.keys() == UnwrappedInputSchema.model_fields.keys()
    assert NestedModel.model_fields.keys() == UnwrappedOutputSchema.model_fields.keys()

    # should not raise
    AbstractInputSchema.model_validate({"inputs": testinput_abstract})
    AbstractOutputSchema.model_validate(testinput_abstract)


def test_create_jacobian_schema():
    arr = testinput["testdiffarr"]
    testoutput = {
        "testdiffarr": {
            "testdiffarr": {
                "object_type": "array",
                "shape": (*arr["shape"], *arr["shape"]),
                "dtype": arr["dtype"],
                "data": {
                    "encoding": "json",
                    "buffer": np.zeros((*arr["shape"], *arr["shape"])).tolist(),
                },
            },
        },
    }

    InputSchema, OutputSchema = create_autodiff_schema(
        NestedModel, NestedModel, "jacobian"
    )

    valid_jac_input = {
        "inputs": testinput,
        "jac_inputs": ["testdiffarr"],
        "jac_outputs": ["testdiffarr"],
    }

    InputSchema.model_validate(valid_jac_input)

    for var in ("jac_inputs", "jac_outputs"):
        jac_testinput = valid_jac_input.copy()

        jac_testinput[var] = [
            "testdiffarr",
            "testfoo.[0].bar.[0]",
            "testbar.{hey there}.bar.[0]",
        ]
        InputSchema.model_validate(jac_testinput)

        with pytest.raises(ValidationError):
            jac_testinput[var] = []
            InputSchema.model_validate(jac_testinput)

        with pytest.raises(ValidationError):
            # non-differentiable array
            jac_testinput[var] = ["testbaz"]
            InputSchema.model_validate(jac_testinput)

        with pytest.raises(ValidationError):
            # non-existing
            jac_testinput[var] = ["moo"]
            InputSchema.model_validate(jac_testinput)

    jac_testinput = valid_jac_input.copy()
    # Test LookupError (from IndexError) (only raised on jac_inputs)
    with pytest.raises(ValidationError):
        jac_testinput["jac_inputs"] = ["testfoo.[2].bar.[0]"]
        InputSchema.model_validate(jac_testinput)

    with pytest.raises(ValidationError):
        jac_testinput["jac_inputs"] = ["testlazysequence.[0].[2]"]
        InputSchema.model_validate(jac_testinput)

    # Test LookupError (from KeyError) (only raised on jac_inputs)
    with pytest.raises(ValidationError):
        jac_testinput["jac_inputs"] = ["testbar.{hey ther}.bar.[0]"]
        InputSchema.model_validate(jac_testinput)

    ctx = {
        "input_keys": valid_jac_input["jac_inputs"],
        "output_keys": valid_jac_input["jac_outputs"],
    }

    OutputSchema.model_validate(testoutput, context=ctx)

    # Input instead of output
    with pytest.raises(ValidationError):
        OutputSchema.model_validate(testinput, context=ctx)

    # Extra keys
    with pytest.raises(ValidationError):
        OutputSchema.model_validate({**testoutput, "foo": 1}, context=ctx)

    # Extra keys (nested)
    with pytest.raises(ValidationError):
        out = testoutput.copy()
        out["testdiffarr"]["foo"] = 1
        OutputSchema.model_validate(out, context=ctx)

    # Missing keys
    with pytest.raises(ValidationError):
        OutputSchema.model_validate({**testoutput, "testdiffarr": {}}, context=ctx)

    # Flat dict
    with pytest.raises(ValidationError):
        OutputSchema.model_validate(testoutput["testdiffarr"], context=ctx)


def test_create_jvp_schema():
    testoutput = {
        "testdiffarr": testinput["testdiffarr"],
    }
    InputSchema, OutputSchema = create_autodiff_schema(NestedModel, NestedModel, "jvp")

    valid_jvp_input = {
        "inputs": testinput,
        "jvp_inputs": ["testdiffarr"],
        "jvp_outputs": ["testdiffarr"],
        "tangent_vector": {"testdiffarr": testinput["testdiffarr"]},
    }

    InputSchema.model_validate(valid_jvp_input)

    for var in ("jvp_inputs", "jvp_outputs"):
        jvp_testinput = valid_jvp_input.copy()

        jvp_testinput[var] = [
            "testdiffarr",
            "testfoo.[0].bar.[0]",
            "testbar.{hey there}.bar.[0]",
            "testlazysequence.[0].[1]",
        ]
        if var == "jvp_inputs":
            jvp_testinput["tangent_vector"] = {
                **jvp_testinput["tangent_vector"],
                "testfoo.[0].bar.[0]": testinput["testfoo"][0]["bar"][0],
                "testbar.{hey there}.bar.[0]": testinput["testbar"]["hey there"]["bar"][
                    0
                ],
                "testlazysequence.[0].[1]": testinput["testlazysequence"][0][1],
            }
        InputSchema.model_validate(jvp_testinput)

        with pytest.raises(ValidationError):
            jvp_testinput[var] = []
            if var == "jvp_inputs":
                jvp_testinput["tangent_vector"] = {}
            InputSchema.model_validate(jvp_testinput)

        with pytest.raises(ValidationError):
            # non-differentiable array
            jvp_testinput[var] = ["testbaz"]
            if var == "jvp_inputs":
                jvp_testinput["tangent_vector"] = {"testbaz": testinput["testbaz"]}
            InputSchema.model_validate(jvp_testinput)

        with pytest.raises(ValidationError):
            # non-existing
            jvp_testinput[var] = ["moo"]
            if var == "jvp_inputs":
                jvp_testinput["tangent_vector"] = {"moo": testinput["testdiffarr"]}
            InputSchema.model_validate(jvp_testinput)

    jvp_testinput = valid_jvp_input.copy()
    # Test LookupError (from IndexError) (only raised on jvp_inputs)
    with pytest.raises(ValidationError):
        jvp_testinput["jvp_inputs"] = ["testfoo.[2].bar.[0]"]
        jvp_testinput["tangent_vector"] = {
            "testfoo.[2].bar.[0]": testinput["testdiffarr"]
        }
        InputSchema.model_validate(jvp_testinput)

    with pytest.raises(ValidationError):
        jvp_testinput["jvp_inputs"] = ["testlazysequence.[0].[2]"]
        jvp_testinput["tangent_vector"] = {
            "testlazysequence.[0].[2]": testinput["testdiffarr"]
        }
        InputSchema.model_validate(jvp_testinput)

    # Test LookupError (from KeyError) (only raised on jvp_inputs)
    with pytest.raises(ValidationError):
        jvp_testinput["jvp_inputs"] = ["testbar.{hey ther}.bar.[0]"]
        jvp_testinput["tangent_vector"] = {
            "testbar.{hey ther}.bar.[0]": testinput["testbar"]["hey there"]["bar"][0]
        }
        InputSchema.model_validate(jvp_testinput)

    # Mis-aligned tangent vector
    jvp_testinput = valid_jvp_input.copy()
    with pytest.raises(ValidationError):
        jvp_testinput["tangent_vector"] = {**jvp_testinput["tangent_vector"], "foo": 1}
        InputSchema.model_validate(jvp_testinput)

    OutputSchema.model_validate(
        testoutput, context={"output_keys": valid_jvp_input["jvp_outputs"]}
    )

    with pytest.raises(ValidationError):
        OutputSchema.model_validate(
            testinput, context={"output_keys": valid_jvp_input["jvp_outputs"]}
        )

    # Extra keys
    with pytest.raises(ValidationError):
        OutputSchema.model_validate(
            {**testoutput, "foo": 1},
            context={"output_keys": valid_jvp_input["jvp_outputs"]},
        )

    # Missing keys
    with pytest.raises(ValidationError):
        OutputSchema.model_validate(
            {}, context={"output_keys": valid_jvp_input["jvp_outputs"]}
        )


def test_create_vjp_schema():
    testoutput = {
        "testdiffarr": testinput["testdiffarr"],
    }

    InputSchema, OutputSchema = create_autodiff_schema(NestedModel, NestedModel, "vjp")

    valid_vjp_input = {
        "inputs": testinput,
        "vjp_inputs": ["testdiffarr"],
        "vjp_outputs": ["testdiffarr"],
        "cotangent_vector": {"testdiffarr": testinput["testdiffarr"]},
    }

    InputSchema.model_validate(valid_vjp_input)

    for var in ("vjp_inputs", "vjp_outputs"):
        vjp_testinput = valid_vjp_input.copy()

        vjp_testinput[var] = [
            "testdiffarr",
            "testfoo.[0].bar.[0]",
            "testlazysequence.[0].[1]",
        ]
        if var == "vjp_outputs":
            vjp_testinput["cotangent_vector"] = {
                **vjp_testinput["cotangent_vector"],
                "testfoo.[0].bar.[0]": 1,
                "testlazysequence.[0].[1]": testinput["testlazysequence"][0][1],
            }
        InputSchema.model_validate(vjp_testinput)

        with pytest.raises(ValidationError):
            vjp_testinput[var] = []
            if var == "vjp_outputs":
                vjp_testinput["cotangent_vector"] = {}
            InputSchema.model_validate(vjp_testinput)

        with pytest.raises(ValidationError):
            # non-differentiable array
            vjp_testinput[var] = ["testbaz"]
            if var == "vjp_outputs":
                vjp_testinput["cotangent_vector"] = {"testbaz": testinput["testbaz"]}
            InputSchema.model_validate(vjp_testinput)

        with pytest.raises(ValidationError):
            # non-existing
            vjp_testinput[var] = ["moo"]
            if var == "vjp_outputs":
                vjp_testinput["cotangent_vector"] = {"moo": 1}
            InputSchema.model_validate(vjp_testinput)

    vjp_testinput = valid_vjp_input.copy()
    # Test LookupError (from IndexError) (only raised on vjp_inputs)
    with pytest.raises(ValidationError):
        vjp_testinput["vjp_inputs"] = ["testfoo.[2].bar.[0]"]
        InputSchema.model_validate(vjp_testinput)

    with pytest.raises(ValidationError):
        vjp_testinput["vjp_inputs"] = ["testlazysequence.[0].[2]"]
        InputSchema.model_validate(vjp_testinput)

    # Test LookupError (from KeyError) (only raised on vjp_inputs)
    with pytest.raises(ValidationError):
        vjp_testinput["vjp_inputs"] = ["testbar.{hey ther}.bar.[0]"]
        InputSchema.model_validate(vjp_testinput)

    # Mis-aligned cotangent vector
    vjp_testinput = valid_vjp_input.copy()
    with pytest.raises(ValidationError):
        vjp_testinput["cotangent_vector"] = {
            **vjp_testinput["cotangent_vector"],
            "foo": 1,
        }
        InputSchema.model_validate(vjp_testinput)

    OutputSchema.model_validate(
        testoutput, context={"input_keys": valid_vjp_input["vjp_inputs"]}
    )

    with pytest.raises(ValidationError):
        OutputSchema.model_validate(
            testinput, context={"input_keys": valid_vjp_input["vjp_inputs"]}
        )

    # Extra keys
    with pytest.raises(ValidationError):
        OutputSchema.model_validate(
            {**testoutput, "foo": 1},
            context={"input_keys": valid_vjp_input["vjp_inputs"]},
        )

    # Missing keys
    with pytest.raises(ValidationError):
        OutputSchema.model_validate(
            {}, context={"input_keys": valid_vjp_input["vjp_inputs"]}
        )


def test_untyped_container_schema_generation():
    # Test with some stuff that's legal in Pydantic but may break the schema generation
    class WeirdModel(BaseModel):
        generic_dict: dict
        generic_list: list
        generic_tuple: tuple
        generic_set: set

    InputSchema, _ = create_apply_schema(WeirdModel, WeirdModel)

    InputSchema.model_validate(
        {
            "inputs": {
                "generic_dict": {"a": 1, "b": "c"},
                "generic_list": [1, 2, 3],
                "generic_tuple": (1, 2, 3),
                "generic_set": {1, 2, 3},
            },
        }
    )


def test_recursive_model():
    class RecursiveModel(BaseModel):
        foo: int
        bar: Optional["RecursiveModel"] | None

    valid_data = {"foo": 1, "bar": {"foo": 2, "bar": {"foo": 3, "bar": None}}}

    # valid pydantic model
    RecursiveModel.model_validate(valid_data)

    # not valid as Tesseract schema
    with pytest.raises(ValueError, match="Recursive model definition"):
        create_apply_schema(RecursiveModel, RecursiveModel)


@pytest.mark.parametrize("endpoint", ["apply", "abstract_eval", "autodiff"])
def test_fancy_pydantic_model(endpoint):
    # Define a model with some fancy features
    from pydantic import AfterValidator, Field, computed_field, model_validator

    def _array_validator(arr):
        # Sharp edge: in abstract eval, this field won't be a NumPy array
        # so we just pass it through
        if not isinstance(arr, np.ndarray):
            return arr
        assert arr.sum() < 2
        return arr

    class InputSchema(BaseModel):
        someint: int = Field(gt=0)

        myarray: Annotated[
            Differentiable[Array[(None,), Float64]], AfterValidator(_array_validator)
        ]

        # All those also include fail-safes for `abstract_eval`
        @computed_field
        @property
        def array_sum(self) -> float:
            if not isinstance(self.myarray, np.ndarray):
                return 0.0
            return float(self.myarray.sum())

        @property
        def array_sum2(self) -> float:
            if not isinstance(self.myarray, np.ndarray):
                return 0.0
            return float(self.myarray.sum())

        @model_validator(mode="after")
        def validate_model(self):
            if not isinstance(self.myarray, np.ndarray):
                return self
            assert self.myarray.sum() + self.someint < 10
            return self

    valid_data = {
        "someint": 1,
        "myarray": {
            "object_type": "array",
            "shape": [5],
            "dtype": "float64",
            "data": {"encoding": "json", "buffer": [0.1, 0.2, 0.3, 0.4, 0.5]},
        },
    }

    # Validate that InputSchema is a valid pydantic model
    InputSchema.model_validate(valid_data)

    # Test that the schema generation works as expected
    if endpoint == "apply":
        ApplyInputSchema, ApplyOutputSchema = create_apply_schema(
            InputSchema, InputSchema
        )

        parsed_inputs = ApplyInputSchema.model_validate({"inputs": valid_data}).inputs

        roundtrip = parsed_inputs.model_dump(mode="json")
        array_sum = roundtrip.pop("array_sum")
        assert array_sum == parsed_inputs.myarray.sum()
        assert roundtrip == valid_data

        # also check output schema
        parsed_outputs = ApplyOutputSchema.model_validate(valid_data)
        assert "array_sum" in parsed_outputs.model_dump()

        # trigger field validator
        with pytest.raises(ValidationError, match="greater_than"):
            ApplyInputSchema.model_validate(
                {
                    "inputs": {
                        "someint": -1,
                        "myarray": valid_data["myarray"],
                    }
                }
            )

        with pytest.raises(ValidationError, match="sum"):
            ApplyInputSchema.model_validate(
                {
                    "inputs": {
                        "someint": 1,
                        "myarray": {
                            "object_type": "array",
                            "shape": [5],
                            "dtype": "float64",
                            "data": {"encoding": "json", "buffer": [1, 1, 1, 1, 1]},
                        },
                    }
                }
            )

        # trigger model validator
        with pytest.raises(ValidationError, match="sum"):
            ApplyInputSchema.model_validate(
                {
                    "inputs": {
                        "someint": 10,
                        "myarray": valid_data["myarray"],
                    }
                }
            )

    elif endpoint == "abstract_eval":
        # Keep everything but replace with dummy values

        # Invalid for abstract eval
        valid_data["myarray"].pop("data")
        valid_data["myarray"].pop("object_type")

        AbstractEvalSchema, _ = create_abstract_eval_schema(InputSchema, InputSchema)
        parsed_inputs = AbstractEvalSchema.model_validate({"inputs": valid_data}).inputs

        roundtrip = parsed_inputs.model_dump(mode="json")
        array_sum = roundtrip.pop("array_sum")
        assert array_sum == 0.0
        assert roundtrip == valid_data

        # trigger field validator on non-array arg
        with pytest.raises(ValidationError, match="greater_than"):
            AbstractEvalSchema.model_validate(
                {
                    "inputs": {
                        "someint": -1,
                        "myarray": valid_data["myarray"],
                    }
                }
            )

    elif endpoint == "autodiff":
        AutodiffSchema, _ = create_autodiff_schema(InputSchema, InputSchema, "jacobian")
        parsed_inputs = AutodiffSchema.model_validate(
            {
                "inputs": valid_data,
                "jac_inputs": ["myarray"],
                "jac_outputs": ["myarray"],
            }
        ).inputs

        roundtrip = parsed_inputs.model_dump(mode="json")
        array_sum = roundtrip.pop("array_sum")
        assert array_sum == parsed_inputs.myarray.sum()
        assert roundtrip == valid_data

        # trigger field validator
        with pytest.raises(ValidationError, match="greater_than"):
            AutodiffSchema.model_validate(
                {
                    "inputs": {
                        "someint": -1,
                        "myarray": valid_data["myarray"],
                    },
                    "jac_inputs": ["myarray"],
                    "jac_outputs": ["myarray"],
                }
            )

        with pytest.raises(ValidationError, match="sum"):
            AutodiffSchema.model_validate(
                {
                    "inputs": {
                        "someint": 1,
                        "myarray": {
                            "object_type": "array",
                            "shape": [5],
                            "dtype": "float64",
                            "data": {"encoding": "json", "buffer": [1, 1, 1, 1, 1]},
                        },
                    },
                    "jac_inputs": ["myarray"],
                    "jac_outputs": ["myarray"],
                }
            )

        # trigger model validator
        with pytest.raises(ValidationError, match="sum"):
            AutodiffSchema.model_validate(
                {
                    "inputs": {
                        "someint": 10,
                        "myarray": valid_data["myarray"],
                    },
                    "jac_inputs": ["myarray"],
                    "jac_outputs": ["myarray"],
                }
            )


@pytest.mark.parametrize(
    "endpoint", ["apply", "abstract_eval", "jacobian", "jvp", "vjp"]
)
def test_json_schema(endpoint):
    if endpoint == "apply":
        InputSchema, OutputSchema = create_apply_schema(NestedModel, NestedModel)
    elif endpoint == "abstract_eval":
        InputSchema, OutputSchema = create_abstract_eval_schema(
            NestedModel, NestedModel
        )
    else:
        InputSchema, OutputSchema = create_autodiff_schema(
            NestedModel, NestedModel, endpoint
        )

    schema_inputs = InputSchema.model_json_schema()
    schema_outputs = OutputSchema.model_json_schema()

    # Test that the JSON schema is valid JSON
    json.dumps(schema_inputs)
    json.dumps(schema_outputs)


def test_model_config_extra_forbid():
    class Child(BaseModel):
        x: str
        model_config: ConfigDict = ConfigDict(extra="allow")

    class Parent(BaseModel):
        child: Child

    ApplyParent = apply_function_to_model_tree(
        Parent,
        lambda x, y: x,
        default_model_config=dict(extra="forbid"),
    )
    ApplyChild = ApplyParent.model_fields["child"].annotation
    assert ApplyChild.model_config["extra"] == "allow"
    assert ApplyParent.model_config["extra"] == "forbid"

    ApplyParent.model_validate({"child": {"x": "foo"}})
    ApplyParent.model_validate({"child": {"x": "foo", "extra": 1}})

    with pytest.raises(ValidationError):
        ApplyParent.model_validate({"child": {"x": "foo"}, "extra": 1})



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import shutil
from pathlib import Path

from pydantic import BaseModel

from tesseract_core.runtime.config import get_config
from tesseract_core.runtime.experimental import (
    InputFileReference,
    OutputFileReference,
)


class InputSchema(BaseModel):
    data: list[InputFileReference]


class OutputSchema(BaseModel):
    data: list[OutputFileReference]


def apply(inputs: InputSchema) -> OutputSchema:
    output_path = Path(get_config().output_path)
    files = []
    for source in inputs.data:
        # source is a pathlib.Path starting with /path/to/input_path/...
        target = output_path / source.name
        # target must be a pathlib.Path at /path/to/output_path
        target = target.with_suffix(".copy")
        shutil.copy(source, target)
        files.append(target)
    return OutputSchema(data=files)



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import importlib
import sys
from pathlib import Path

# Recursive import of the whole package to ensure all deps are present
try:
    for path in Path(__file__).parent.glob("**/*.py"):
        if path.stem == "__init__":
            continue
        if path.stem.startswith("app_"):
            continue
        if not (path.parent / "__init__.py").exists():
            continue

        package_path = ".".join(
            path.relative_to(Path(__file__).parent).with_suffix("").parts
        )
        importlib.import_module(f".{package_path}", __package__)
except ModuleNotFoundError as e:
    print(
        f"Failed to import {path}: {e}\n\n",
        "You are likely missing dependencies. "
        "Try running the following command to install runtime dependencies:\n\n"
        "   $ pip install tesseract-core[runtime]\n",
        file=sys.stderr,
    )
    sys.exit(1)

# Import public API
from .schema_types import (
    Array,
    Differentiable,
    Float16,
    Float32,
    Float64,
    Int8,
    Int16,
    Int32,
    Int64,
    ShapeDType,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
)

__all__ = [
    "Array",
    "Differentiable",
    "Float16",
    "Float32",
    "Float64",
    "Int8",
    "Int16",
    "Int32",
    "Int64",
    "ShapeDType",
    "UInt8",
    "UInt16",
    "UInt32",
    "UInt64",
]



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import base64
import json
import os
import platform
import signal
import subprocess
import sys
import time
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from textwrap import dedent

import numpy as np
import pytest
import requests
from fastapi.testclient import TestClient

from tesseract_core.runtime.serve import create_rest_api

test_input = {
    "a": [1.0, 2.0, 3.0],
    "b": [1, 1, 1],
    "s": 2.5,
}


def is_wsl():
    """Check if the current environment is WSL."""
    kernel = platform.uname().release
    return "Microsoft" in kernel or "WSL" in kernel


def array_from_json(json_data):
    encoding = json_data["data"]["encoding"]
    if encoding == "base64":
        decoded_buffer = base64.b64decode(json_data["data"]["buffer"])
        array = np.frombuffer(decoded_buffer, dtype=json_data["dtype"]).reshape(
            json_data["shape"]
        )
    elif encoding == "json":
        array = np.array(json_data["data"]["buffer"], dtype=json_data["dtype"]).reshape(
            json_data["shape"]
        )

    return array


def model_to_json(model):
    return json.loads(model.model_dump_json())


@contextmanager
def serve_in_subprocess(api_file, port, num_workers=1, timeout=30.0):
    try:
        proc = subprocess.Popen(
            [
                sys.executable,
                "-c",
                "from tesseract_core.runtime.serve import serve; "
                f"serve(host='localhost', port={port}, num_workers={num_workers})",
            ],
            env=dict(os.environ, TESSERACT_API_PATH=api_file),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # wait for server to start
        while True:
            try:
                response = requests.get(f"http://localhost:{port}/health")
            except requests.exceptions.ConnectionError:
                pass
            else:
                if response.status_code == 200:
                    break

            time.sleep(0.1)
            timeout -= 0.1

            if timeout < 0:
                raise TimeoutError("Server did not start in time")

        yield f"http://localhost:{port}"

    finally:
        proc.send_signal(signal.SIGINT)
        stdout, stderr = proc.communicate()
        print(stdout.decode())
        print(stderr.decode())
        proc.wait(timeout=5)


@pytest.fixture
def http_client(dummy_tesseract_module):
    """A test HTTP client."""
    rest_api = create_rest_api(dummy_tesseract_module)
    return TestClient(rest_api)


@pytest.mark.parametrize(
    "format",
    [
        "json",
        "json+base64",
        pytest.param("json+binref", marks=pytest.mark.xfail),  # FIXME
        "msgpack",
    ],
)
def test_create_rest_api_apply_endpoint(http_client, dummy_tesseract_module, format):
    """Test we can get an Apply endpoint from generated API."""
    test_inputs = dummy_tesseract_module.InputSchema.model_validate(test_input)

    response = http_client.post(
        "/apply",
        json={"inputs": model_to_json(test_inputs)},
        headers={"Accept": f"application/{format}"},
    )

    assert response.status_code == 200, response.text

    if format in {"json", "json+base64"}:
        result = array_from_json(response.json()["result"])
        assert np.array_equal(result, np.array([3.5, 6.0, 8.5]))
    elif format == "msgpack":
        assert (
            response.content
            == b"\x81\xa6result\x85\xc4\x02nd\xc3\xc4\x04type\xa3<f4\xc4\x04kind\xc4\x00\xc4\x05shape\x91\x03\xc4\x04data\xc4\x0c\x00\x00`@\x00\x00\xc0@\x00\x00\x08A"  # noqa: E501
        )
    elif format == "json+binref":
        raise NotImplementedError()


def test_create_rest_api_jacobian_endpoint(http_client, dummy_tesseract_module):
    """Test we can get a Jacobian endpoint from generated API."""
    test_inputs = dummy_tesseract_module.InputSchema.model_validate(test_input)

    response = http_client.post(
        "/jacobian",
        json={
            "inputs": model_to_json(test_inputs),
            "jac_inputs": ["a", "b"],
            "jac_outputs": ["result"],
        },
    )

    assert response.status_code == 200, response.text
    result = response.json()
    expected = dummy_tesseract_module.jacobian(test_inputs, {"a", "b"}, {"result"})

    assert result.keys() == expected.keys()
    assert np.array_equal(
        array_from_json(result["result"]["a"]), expected["result"]["a"]
    )


def test_create_rest_api_generates_health_endpoint(http_client):
    """Test we can get health endpoint from generated API."""
    response = http_client.get("/health")
    assert response.json() == {"status": "ok"}


def test_get_input_schema(http_client):
    response = http_client.get("/input_schema")

    assert response.status_code == 200, response.text


def test_get_output_schema(http_client):
    response = http_client.get("/output_schema")

    assert response.status_code == 200, response.text


def test_post_abstract_eval(http_client):
    payload = {
        "inputs": {
            "a": {"dtype": "float64", "shape": [4]},
            "b": {"dtype": "float64", "shape": [4]},
            "s": 1.0,
            "normalize": False,
        }
    }
    response = http_client.post("/abstract_eval", json=payload)

    assert response.status_code == 200, response.text
    assert response.json() == {"result": {"shape": [4], "dtype": "float64"}}


def test_post_abstract_eval_throws_validation_errors(http_client):
    response = http_client.post("/abstract_eval", json={"what": {"is": "this"}})

    assert response.status_code == 422, response.text
    errors = response.json()["detail"]
    error_types = [e["type"] for e in errors]

    assert "missing" in error_types
    assert "extra_forbidden" in error_types


def test_get_openapi_schema(http_client):
    response = http_client.get("/openapi.json")

    assert response.status_code == 200, response.text
    assert response.json()["info"]["title"] == "Tesseract"
    assert response.json()["paths"]


@pytest.mark.skipif(
    is_wsl(),
    reason="flaky on Windows",
)
def test_threading_sanity(tmpdir, free_port):
    """Test with a Tesseract that requires to be run in the main thread.

    This is important so we don't require users to be aware of threading issues.
    """
    TESSERACT_API = dedent(
        """
    import threading
    from pydantic import BaseModel

    assert threading.current_thread() == threading.main_thread()

    class InputSchema(BaseModel):
        pass

    class OutputSchema(BaseModel):
        pass

    def apply(input: InputSchema) -> OutputSchema:
        assert threading.current_thread() == threading.main_thread()
        return OutputSchema()
    """
    )

    api_file = tmpdir / "tesseract_api.py"

    with open(api_file, "w") as f:
        f.write(TESSERACT_API)

    # We can't run the server in the same process because it will use threading under the hood
    # so we need to spawn a new process instead
    with serve_in_subprocess(api_file, free_port) as url:
        response = requests.post(f"{url}/apply", json={"inputs": {}})
        assert response.status_code == 200, response.text


@pytest.mark.skipif(
    is_wsl(),
    reason="flaky on Windows",
)
def test_multiple_workers(tmpdir, free_port):
    """Test that the server can be run with multiple worker processes."""
    TESSERACT_API = dedent(
        """
    import time
    import multiprocessing
    from pydantic import BaseModel

    class InputSchema(BaseModel):
        pass

    class OutputSchema(BaseModel):
        pid: int

    def apply(input: InputSchema) -> OutputSchema:
        return OutputSchema(pid=multiprocessing.current_process().pid)
    """
    )

    api_file = tmpdir / "tesseract_api.py"

    with open(api_file, "w") as f:
        f.write(TESSERACT_API)

    with serve_in_subprocess(api_file, free_port, num_workers=2) as url:
        # Fire back-to-back requests to the server and check that they are handled
        # by different workers (i.e. different PIDs)
        post_request = lambda _: requests.post(f"{url}/apply", json={"inputs": {}})

        with ThreadPoolExecutor(max_workers=4) as executor:
            # Fire a lot of requests in parallel
            futures = executor.map(post_request, range(100))
            responses = list(futures)

        # Check that all responses are 200
        for response in responses:
            assert response.status_code == 200, response.text

        # Check that not all pids are the same
        # (i.e. the requests were handled by different workers)
        pids = set(response.json()["pid"] for response in responses)
        assert len(pids) > 1, "All requests were handled by the same worker"


def test_debug_mode(dummy_tesseract_module, monkeypatch):
    import tesseract_core.runtime.config
    from tesseract_core.runtime.config import update_config

    def apply_that_raises(inputs):
        raise ValueError("This is a test error")

    orig_config = tesseract_core.runtime.config._current_config

    monkeypatch.setattr(dummy_tesseract_module, "apply", apply_that_raises)

    try:
        update_config(debug=False, api_path=dummy_tesseract_module.__file__)
        rest_api = create_rest_api(dummy_tesseract_module)
        http_client = TestClient(rest_api, raise_server_exceptions=False)

        response = http_client.post(
            "/apply",
            json={
                "inputs": model_to_json(
                    dummy_tesseract_module.InputSchema.model_validate(test_input)
                )
            },
        )
        assert response.status_code == 500, response.text
        assert response.text == "Internal Server Error"
        assert "This is a test error" not in response.text
    finally:
        tesseract_core.runtime.config._current_config = orig_config

    try:
        update_config(debug=True, api_path=dummy_tesseract_module.__file__)
        rest_api = create_rest_api(dummy_tesseract_module)
        http_client = TestClient(rest_api, raise_server_exceptions=False)

        response = http_client.post(
            "/apply",
            json={
                "inputs": model_to_json(
                    dummy_tesseract_module.InputSchema.model_validate(test_input)
                )
            },
        )
        assert response.status_code == 500, response.text
        assert "This is a test error" in response.text
        assert "Traceback" in response.text
    finally:
        tesseract_core.runtime.config._current_config = orig_config



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

# Tesseract API module for {{name}}
# Generated by tesseract {{version}} on {{timestamp}}


from pydantic import BaseModel

#
# Schemas
#


class InputSchema(BaseModel):
    pass


class OutputSchema(BaseModel):
    pass


#
# Required endpoints
#


def apply(inputs: InputSchema) -> OutputSchema:
    return OutputSchema()


#
# Optional endpoints
#

# import numpy as np

# def jacobian(inputs: InputSchema, jac_inputs: set[str], jac_outputs: set[str]):
#     return {}

# def jacobian_vector_product(
#     inputs: InputSchema,
#     jvp_inputs: set[str],
#     jvp_outputs: set[str],
#     tangent_vector: dict[str, np.typing.ArrayLike]
# ) -> dict[str, np.typing.ArrayLike]:
#     return {}

# def vector_jacobian_product(
#     inputs: InputSchema,
#     vjp_inputs: set[str],
#     vjp_outputs: set[str],
#     cotangent_vector: dict[str, np.typing.ArrayLike]
# ) -> dict[str, np.typing.ArrayLike]:
#     return {}

# def abstract_eval(abstract_inputs):
#     return {}



from pathlib import Path

from rich import print

from tesseract_core import Tesseract

here = Path(__file__).parent.resolve()
input_path = Path("./testdata")
output_path = Path("./output")

# these are relative to input_path
data = [
    "sample_0.json",
    "sample_1.json",
    "sample_2.json",
    "sample_3.json",
    "sample_4.json",
    "sample_5.json",
    "sample_6.json",
    "sample_7.json",
    "sample_8.json",
    "sample_9.json",
]

with Tesseract.from_tesseract_api(
    "tesseract_api.py", input_path=input_path, output_path=output_path
) as tess:
    result = tess.apply({"data": data})
    print(result)
    assert all((output_path / p).exists() for p in result["data"])


with Tesseract.from_image(
    "filereference",
    input_path=input_path,
    output_path=output_path,
) as tess:
    result = tess.apply({"data": data})
    print(result)
    assert all((output_path / p).exists() for p in result["data"])



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import os
import shlex
import shutil
from collections.abc import Sequence
from pathlib import Path
from typing import Annotated, Any

from pydantic import BaseModel, BeforeValidator, ConfigDict


def validate_executable(value: str | Sequence[str]) -> tuple[str, ...]:
    """Get the path to the requested program."""
    if isinstance(value, str):
        exe, *args = shlex.split(value)
    else:
        exe, *args = value

    exe = shutil.which(exe)
    if exe is None:
        raise ValueError(f"Executable `{value}` not found.")

    exe_path = Path(exe)
    if not exe_path.is_file():
        raise ValueError(f"{exe} is not a file.")
    if not os.access(exe_path, os.X_OK):
        raise ValueError(f"{exe} is not executable.")

    return (str(exe_path.resolve()), *args)


def maybe_split_args(value: str | Sequence[str]) -> tuple[str, ...]:
    """Split arguments if they are passed as a string."""
    if isinstance(value, str):
        value = shlex.split(value)
    return tuple(value)


class RuntimeConfig(BaseModel):
    """Available runtime configuration."""

    docker_executable: Annotated[
        tuple[str, ...], BeforeValidator(validate_executable)
    ] = ("docker",)
    docker_compose_executable: Annotated[
        tuple[str, ...], BeforeValidator(validate_executable)
    ] = ("docker", "compose")
    docker_build_args: Annotated[
        tuple[str, ...], BeforeValidator(maybe_split_args)
    ] = ()
    docker_run_args: Annotated[tuple[str, ...], BeforeValidator(maybe_split_args)] = ()

    model_config = ConfigDict(frozen=True, extra="forbid")


def update_config(**kwargs: Any) -> None:
    """Create a new runtime configuration from the current environment.

    Passed keyword arguments will override environment variables.
    """
    global _current_config

    conf_settings = {}
    for field in RuntimeConfig.model_fields.keys():
        env_key = f"TESSERACT_{field.upper()}"
        if env_key in os.environ:
            conf_settings[field] = os.environ[env_key]

    for field in _config_overrides:
        conf_settings[field] = getattr(_current_config, field)

    conf_settings.update(kwargs)
    config = RuntimeConfig(**conf_settings)

    _config_overrides.update(set(conf_settings.keys()))
    _current_config = config


_current_config = None
_config_overrides = set()


def get_config() -> RuntimeConfig:
    """Return the current runtime configuration."""
    if _current_config is None:
        update_config()
    assert _current_config is not None
    return _current_config



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Re-generate test data for this Tesseract."""

from pathlib import Path

import numpy as np
from pydantic import RootModel

from tesseract_core.runtime import Array, Float32

rng = np.random.default_rng(0)

OUTDIR = Path(__file__).parent / "testdata"


class TestArray(RootModel):
    root: Array[(None, 3), Float32]


if __name__ == "__main__":
    OUTDIR.mkdir(parents=True, exist_ok=True)
    binref_uuid = "a2af4236-15c5-40ec-9014-b79f9cbad50c"
    (OUTDIR / f"{binref_uuid}.bin").write_bytes(b"")

    for idx in range(10):
        if idx < 9:
            data = rng.random((3, 3))
        else:
            # Create one sample with a different shape
            data = rng.random((4, 3))

        # Create some arrays of all supported encodings
        if idx % 3 == 0:
            encoding = "json"
        elif idx % 3 == 1:
            encoding = "base64"
        else:
            encoding = "binref"

        out_json = TestArray(data).model_dump_json(
            indent=2,
            context={
                "array_encoding": encoding,
                "base_dir": OUTDIR,
                "__binref_uuid": binref_uuid,
            },
        )

        with open(OUTDIR / f"sample_{idx}.json", "w") as f:
            f.write(out_json)



#!/usr/bin/env python

"""Support for serialization of numpy data types with msgpack."""

# Copyright (c) 2013-2022, Lev E. Givon
# All rights reserved.
# Distributed under the terms of the BSD license:
# http://www.opensource.org/licenses/bsd-license

import functools
import pickle
import sys
import warnings

import msgpack
import numpy as np
from msgpack import Packer as _Packer
from msgpack import Unpacker as _Unpacker
from msgpack import unpack as _unpack
from msgpack import unpackb as _unpackb

if sys.version_info >= (3, 0):

    def ndarray_to_bytes(obj):
        if obj.dtype == "O":
            return obj.dumps()
        else:
            if sys.platform == "darwin":
                return obj.tobytes()
            else:
                return obj.data if obj.flags["C_CONTIGUOUS"] else obj.tobytes()

    num_to_bytes = lambda obj: obj.data

    def tostr(x):
        if isinstance(x, bytes):
            return x.decode()
        else:
            return str(x)
else:

    def ndarray_to_bytes(obj):
        if obj.dtype == "O":
            return obj.dumps()
        else:
            if sys.platform == "darwin":
                return obj.tobytes()
            else:
                return (
                    memoryview(obj.data) if obj.flags["C_CONTIGUOUS"] else obj.tobytes()
                )

    num_to_bytes = lambda obj: memoryview(obj.data)

    def tostr(x):
        return x


def encode(obj, chain=None):
    """Data encoder for serializing numpy data types."""
    if isinstance(obj, np.ndarray):
        # If the dtype is structured, store the interface description;
        # otherwise, store the corresponding array protocol type string:
        if obj.dtype.kind in ("V", "O"):
            kind = bytes(obj.dtype.kind, "ascii")
            descr = obj.dtype.descr
        else:
            kind = b""
            descr = obj.dtype.str

        return {
            b"nd": True,
            b"type": descr,
            b"kind": kind,
            b"shape": obj.shape,
            b"data": ndarray_to_bytes(obj),
        }
    elif isinstance(obj, (np.bool_, np.number)):
        return {b"nd": False, b"type": obj.dtype.str, b"data": num_to_bytes(obj)}
    elif isinstance(obj, complex):
        return {b"complex": True, b"data": obj.__repr__()}
    else:
        return obj if chain is None else chain(obj)


def decode(obj, chain=None):
    """Decoder for deserializing numpy data types."""
    try:
        if b"nd" in obj:
            if obj[b"nd"] is True:
                # Check if b'kind' is in obj to enable decoding of data
                # serialized with older versions (#20) or data
                # that had dtype == 'O' (#46):
                if b"kind" in obj and obj[b"kind"] == b"V":
                    descr = [
                        tuple(tostr(t) if type(t) is bytes else t for t in d)
                        for d in obj[b"type"]
                    ]
                elif b"kind" in obj and obj[b"kind"] == b"O":
                    return pickle.loads(obj[b"data"])
                else:
                    descr = obj[b"type"]
                return np.ndarray(
                    buffer=obj[b"data"], dtype=_unpack_dtype(descr), shape=obj[b"shape"]
                )
            else:
                descr = obj[b"type"]
                return np.frombuffer(obj[b"data"], dtype=_unpack_dtype(descr))[0]
        elif b"complex" in obj:
            return complex(tostr(obj[b"data"]))
        else:
            return obj if chain is None else chain(obj)
    except KeyError:
        return obj if chain is None else chain(obj)


def _unpack_dtype(dtype):
    """Unpack dtype descr, recursively unpacking nested structured dtypes."""
    if isinstance(dtype, (list, tuple)):
        # Unpack structured dtypes of the form: (name, type, *shape)
        dtype = [
            (subdtype[0], _unpack_dtype(subdtype[1])) + tuple(subdtype[2:])
            for subdtype in dtype
        ]
    return np.dtype(dtype)


if msgpack.version < (1, 0, 0):
    warnings.warn(
        "support for msgpack < 1.0.0 will be removed in a future release",
        DeprecationWarning,
    )

# Maintain support for msgpack < 0.4.0:
if msgpack.version < (0, 4, 0):

    class Packer(_Packer):
        def __init__(
            self,
            default=None,
            encoding="utf-8",
            unicode_errors="strict",
            use_single_float=False,
            autoreset=1,
        ):
            default = functools.partial(encode, chain=default)
            super(Packer, self).__init__(
                default=default,
                encoding=encoding,
                unicode_errors=unicode_errors,
                use_single_float=use_single_float,
                autoreset=autoreset,
            )

    class Unpacker(_Unpacker):
        def __init__(
            self,
            file_like=None,
            read_size=0,
            use_list=None,
            object_hook=None,
            object_pairs_hook=None,
            list_hook=None,
            encoding="utf-8",
            unicode_errors="strict",
            max_buffer_size=0,
        ):
            object_hook = functools.partial(decode, chain=object_hook)
            super(Unpacker, self).__init__(
                file_like=file_like,
                read_size=read_size,
                use_list=use_list,
                object_hook=object_hook,
                object_pairs_hook=object_pairs_hook,
                list_hook=list_hook,
                encoding=encoding,
                unicode_errors=unicode_errors,
                max_buffer_size=max_buffer_size,
            )

elif msgpack.version < (1, 0, 0):

    class Packer(_Packer):
        def __init__(
            self,
            default=None,
            unicode_errors="strict",
            use_single_float=False,
            autoreset=1,
            use_bin_type=True,
            strict_types=False,
        ):
            default = functools.partial(encode, chain=default)
            super(Packer, self).__init__(
                default=default,
                unicode_errors=unicode_errors,
                use_single_float=use_single_float,
                autoreset=autoreset,
                use_bin_type=use_bin_type,
                strict_types=strict_types,
            )

    class Unpacker(_Unpacker):
        def __init__(
            self,
            file_like=None,
            read_size=0,
            use_list=None,
            raw=False,
            object_hook=None,
            object_pairs_hook=None,
            list_hook=None,
            unicode_errors="strict",
            max_buffer_size=0,
            ext_hook=msgpack.ExtType,
        ):
            object_hook = functools.partial(decode, chain=object_hook)
            super(Unpacker, self).__init__(
                file_like=file_like,
                read_size=read_size,
                use_list=use_list,
                raw=raw,
                object_hook=object_hook,
                object_pairs_hook=object_pairs_hook,
                list_hook=list_hook,
                unicode_errors=unicode_errors,
                max_buffer_size=max_buffer_size,
                ext_hook=ext_hook,
            )

else:

    class Packer(_Packer):
        def __init__(
            self,
            default=None,
            use_single_float=False,
            autoreset=True,
            use_bin_type=True,
            strict_types=False,
            datetime=False,
            unicode_errors=None,
        ):
            default = functools.partial(encode, chain=default)
            super(Packer, self).__init__(
                default=default,
                use_single_float=use_single_float,
                autoreset=autoreset,
                use_bin_type=use_bin_type,
                strict_types=strict_types,
                datetime=datetime,
                unicode_errors=unicode_errors,
            )

    class Unpacker(_Unpacker):
        def __init__(
            self,
            file_like=None,
            read_size=0,
            use_list=True,
            raw=False,
            timestamp=0,
            strict_map_key=True,
            object_hook=None,
            object_pairs_hook=None,
            list_hook=None,
            unicode_errors=None,
            max_buffer_size=100 * 1024 * 1024,
            ext_hook=msgpack.ExtType,
            max_str_len=-1,
            max_bin_len=-1,
            max_array_len=-1,
            max_map_len=-1,
            max_ext_len=-1,
        ):
            object_hook = functools.partial(decode, chain=object_hook)
            super(Unpacker, self).__init__(
                file_like=file_like,
                read_size=read_size,
                use_list=use_list,
                raw=raw,
                timestamp=timestamp,
                strict_map_key=strict_map_key,
                object_hook=object_hook,
                object_pairs_hook=object_pairs_hook,
                list_hook=list_hook,
                unicode_errors=unicode_errors,
                max_buffer_size=max_buffer_size,
                ext_hook=ext_hook,
                max_str_len=max_str_len,
                max_bin_len=max_bin_len,
                max_array_len=max_array_len,
                max_map_len=max_map_len,
                max_ext_len=max_ext_len,
            )


def pack(o, stream, **kwargs):
    """Pack an object and write it to a stream."""
    packer = Packer(**kwargs)
    stream.write(packer.pack(o))


def packb(o, **kwargs):
    """Pack an object and return the packed bytes."""
    return Packer(**kwargs).pack(o)


def unpack(stream, **kwargs):
    """Unpack a packed object from a stream."""
    object_hook = kwargs.get("object_hook")
    kwargs["object_hook"] = functools.partial(decode, chain=object_hook)
    return _unpack(stream, **kwargs)


def unpackb(packed, **kwargs):
    """Unpack a packed object."""
    object_hook = kwargs.get("object_hook")
    kwargs["object_hook"] = functools.partial(decode, chain=object_hook)
    return _unpackb(packed, **kwargs)


load = unpack
loads = unpackb
dump = pack
dumps = packb


def patch():
    """Monkey patch msgpack module to enable support for serializing numpy types."""
    msgpack.Packer = Packer
    msgpack.Unpacker = Unpacker
    msgpack.load = unpack
    msgpack.loads = unpackb
    msgpack.dump = pack
    msgpack.dumps = packb
    msgpack.pack = pack
    msgpack.packb = packb
    msgpack.unpack = unpack
    msgpack.unpackb = unpackb



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

# Tesseract API module for {{name}}
# Generated by tesseract {{version}} on {{timestamp}}

from typing import Any

import equinox as eqx
import jax
from pydantic import BaseModel

from tesseract_core.runtime import Differentiable, Float32
from tesseract_core.runtime.tree_transforms import filter_func, flatten_with_paths

#
# Schemata
#


# Note: This template uses equinox filter_jit to automatically treat non-array
# inputs/outputs as static. As Tesseract scalar objects (e.g. Float32) are
# essentially just wrappers around numpy 0D arrays, they will be considered to
# be dynamic and will be traced by JAX.
# If you want to treat scalar numerical values as static you will need to use
# built-in Python types (e.g. float, int) instead of Float32.


class InputSchema(BaseModel):
    example: Differentiable[Float32]


class OutputSchema(BaseModel):
    example: Differentiable[Float32]


#
# Required endpoints
#


# TODO: Add or import your function here, must be JAX-jittable and
# take/return a single pytree as an input/output conforming respectively
# to Input/OutputSchema
@eqx.filter_jit
def apply_jit(inputs: dict) -> dict:
    return inputs


def apply(inputs: InputSchema) -> OutputSchema:
    # Optional: Insert any pre-processing/setup that doesn't require tracing
    # and is only required when specifically running your apply function
    # and not your differentiable endpoints.
    # For example, you might want to set up a logger or mlflow server.
    # Pre-processing should not modify any input that could impact the
    # differentiable outputs in a nonlinear way (a constant shift
    # should be safe)

    out = apply_jit(inputs.model_dump())

    # Optional: Insert any post-processing that doesn't require tracing
    # For example, you might want to save to disk or modify a non-differentiable
    # output. Again, do not modify any differentiable output in a non-linear way.
    return out


#
# Jax-handled AD endpoints (no need to modify)
#


def jacobian(
    inputs: InputSchema,
    jac_inputs: set[str],
    jac_outputs: set[str],
):
    return jac_jit(inputs.model_dump(), tuple(jac_inputs), tuple(jac_outputs))


def jacobian_vector_product(
    inputs: InputSchema,
    jvp_inputs: set[str],
    jvp_outputs: set[str],
    tangent_vector: dict[str, Any],
):
    return jvp_jit(
        inputs.model_dump(),
        tuple(jvp_inputs),
        tuple(jvp_outputs),
        tangent_vector,
    )


def vector_jacobian_product(
    inputs: InputSchema,
    vjp_inputs: set[str],
    vjp_outputs: set[str],
    cotangent_vector: dict[str, Any],
):
    return vjp_jit(
        inputs.model_dump(),
        tuple(vjp_inputs),
        tuple(vjp_outputs),
        cotangent_vector,
    )


def abstract_eval(abstract_inputs):
    """Calculate output shape of apply from the shape of its inputs."""
    is_shapedtype_dict = lambda x: type(x) is dict and (x.keys() == {"shape", "dtype"})
    is_shapedtype_struct = lambda x: isinstance(x, jax.ShapeDtypeStruct)

    jaxified_inputs = jax.tree.map(
        lambda x: jax.ShapeDtypeStruct(**x) if is_shapedtype_dict(x) else x,
        abstract_inputs.model_dump(),
        is_leaf=is_shapedtype_dict,
    )
    dynamic_inputs, static_inputs = eqx.partition(
        jaxified_inputs, filter_spec=is_shapedtype_struct
    )

    def wrapped_apply(dynamic_inputs):
        inputs = eqx.combine(static_inputs, dynamic_inputs)
        return apply_jit(inputs)

    jax_shapes = jax.eval_shape(wrapped_apply, dynamic_inputs)
    return jax.tree.map(
        lambda x: (
            {"shape": x.shape, "dtype": str(x.dtype)} if is_shapedtype_struct(x) else x
        ),
        jax_shapes,
        is_leaf=is_shapedtype_struct,
    )


#
# Helper functions
#


@eqx.filter_jit
def jac_jit(
    inputs: dict,
    jac_inputs: tuple[str],
    jac_outputs: tuple[str],
):
    filtered_apply = filter_func(apply_jit, inputs, jac_outputs)
    return jax.jacrev(filtered_apply)(
        flatten_with_paths(inputs, include_paths=jac_inputs)
    )


@eqx.filter_jit
def jvp_jit(
    inputs: dict, jvp_inputs: tuple[str], jvp_outputs: tuple[str], tangent_vector: dict
):
    filtered_apply = filter_func(apply_jit, inputs, jvp_outputs)
    return jax.jvp(
        filtered_apply,
        [flatten_with_paths(inputs, include_paths=jvp_inputs)],
        [tangent_vector],
    )[1]


@eqx.filter_jit
def vjp_jit(
    inputs: dict,
    vjp_inputs: tuple[str],
    vjp_outputs: tuple[str],
    cotangent_vector: dict,
):
    filtered_apply = filter_func(apply_jit, inputs, vjp_outputs)
    _, vjp_func = jax.vjp(
        filtered_apply, flatten_with_paths(inputs, include_paths=vjp_inputs)
    )
    return vjp_func(cotangent_vector)[0]



# Copyright 2025 Pasteur Labs. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


class UserError(Exception):
    """Exception raised for anything that is the user's fault."""

    pass


class ValidationError(UserError):
    """Exception raised for input validation errors."""

    pass
